import copy
import glob
import os
import sys
import pprint
from itertools import groupby
from textwrap import wrap
import numpy as np
import pandas as pd
import pylab as plt
import tqdm

from .. import haven_jobs as hjb
from .. import haven_utils as hu
from .. import haven_share as hd


def get_score_df(
    exp_list,
    savedir_base,
    filterby_list=None,
    columns=None,
    score_columns=None,
    verbose=True,
    wrap_size=8,
    hparam_diff=0,
    flatten_columns=True,
    show_meta=True,
    show_max_min=True,
    add_prefix=False,
    score_list_name="score_list.pkl",
    in_latex_format=False,
    avg_across=None,
    return_columns=False,
    show_exp_ids=False,
):
    """Get a table showing the scores for the given list of experiments

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    columns : list, optional
        a list of columns you would like to display, by default None
    savedir_base : str, optional
        A directory where experiments are saved

    Returns
    -------
    DataFrame
        a dataframe showing the scores obtained by the experiments

    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base,
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> df = hr.get_score_df(exp_list, savedir_base=savedir_base, columns=['train_loss', 'exp_id'])
    >>> print(df)
    """
    if len(exp_list) == 0:
        if verbose:
            print("exp_list is empty...")
        if return_columns:
            return pd.DataFrame([]), [], []
        else:
            return pd.DataFrame([])
    exp_list = hu.filter_exp_list(exp_list, filterby_list, savedir_base=savedir_base, verbose=verbose)

    # aggregate results
    hparam_list = set()
    result_list = []
    for exp_dict in exp_list:
        result_dict = {}
        exp_id = hu.hash_dict(exp_dict)
        if avg_across is not None:
            tmp_dict = copy.deepcopy(exp_dict)
            del tmp_dict[avg_across]
            result_dict["_" + avg_across] = hu.hash_dict(tmp_dict)

        savedir = os.path.join(savedir_base, exp_id)
        score_list_fname = os.path.join(savedir, score_list_name)
        exp_dict_fname = os.path.join(savedir, "exp_dict.json")

        if flatten_columns:
            exp_dict_flat = hu.flatten_column(exp_dict, flatten_list=True)
        else:
            exp_dict_flat = exp_dict

        hparam_columns = columns or list(exp_dict_flat.keys())
        for hc in hparam_columns:
            hparam_list.add(hc)
        for k in hparam_columns:
            if k == "exp_id":
                continue
            if add_prefix:
                k_new = "(hparam) " + k
            else:
                k_new = k
            if k not in exp_dict_flat:
                continue
            result_dict[k_new] = exp_dict_flat[k]

        if os.path.exists(score_list_fname) and show_meta:
            result_dict["started_at"] = hu.time_to_montreal(exp_dict_fname)
            result_dict["creation_time"] = os.path.getctime(exp_dict_fname)
        else:
            result_dict["creation_time"] = -1

        if show_exp_ids or "exp_id" in hparam_columns:
            result_dict["exp_id"] = exp_id

        # hparam_columns = [k for k in result_dict.keys() if k not in ['creation_time']]

        if not os.path.exists(score_list_fname):
            if verbose:
                print("%s: %s is missing" % (exp_id, score_list_name))

        else:
            try:
                score_list = hu.load_pkl(score_list_fname)
            except Exception:
                print("%s: %s is corrupt" % (exp_id, score_list_name))
            score_df = pd.DataFrame(score_list)
            metric_columns = score_columns or score_df.columns
            if len(score_list):
                for k in metric_columns:
                    if k not in score_df.columns:
                        continue
                    v = np.array(score_df[k])
                    if "float" in str(v.dtype):
                        v = v[~np.isnan(v)]

                    if len(v):
                        if add_prefix:
                            k_new = "(metric) " + k
                        else:
                            k_new = k

                        if "float" in str(v.dtype):
                            result_dict[k_new] = v[-1]
                            if show_max_min:
                                result_dict[k_new + " (max)"] = v.max()
                                result_dict[k_new + " (min)"] = v.min()

                        else:
                            result_dict[k_new] = v[-1]

        result_list += [result_dict]

    # create table
    df = pd.DataFrame(result_list)
    hparam_columns = [c for c in hparam_list if c not in ["exp_id"]]
    metric_columns = [c for c in df.columns if c not in hparam_columns + ["creation_time"]]
    # print(avg_across)
    if avg_across is not None:
        df_avg = df.groupby("_" + avg_across)
        df_tmp = df_avg[hparam_columns].first().join(df_avg[metric_columns].agg([np.mean, np.std]))

        if ("_" + avg_across) in df_tmp.columns:
            del df_tmp["_" + avg_across]

        df_tmp = df_tmp.reset_index()

        if ("_" + avg_across) in df_tmp.columns:
            del df_tmp["_" + avg_across]

        df = df_tmp

    if "creation_time" in df.columns:
        df = df.sort_values(by="creation_time")
        del df["creation_time"]

    # wrap text for prettiness
    df = hu.pretty_print_df(df)

    if (hparam_diff > 0 and len(df) > 1) or columns is None:
        cols = hu.get_diff_columns(df, min_threshold=hparam_diff, max_threshold="auto")
        df = df[cols]

    df = hu.sort_df_columns(df, also_first=hparam_columns)
    if in_latex_format:
        return df.to_latex(index=False)

    if return_columns:
        return df, sorted(hparam_columns), sorted(metric_columns)
    return df


def get_score_lists(
    exp_list, savedir_base, filterby_list=None, verbose=True, score_list_name="score_list.pkl", return_as_dict=False
):
    """[summary]

    Parameters
    ----------
    exp_list : [type]
        A list of experiments, each defines a single set of hyper-parameters
    savedir_base : [type]
        [description]
    filterby_list : [type], optional
        [description], by default None

    Returns
    -------
    [type]
        [description]
    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base,
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> lists_of_score_lists = hr.get_score_lists(exp_list, savedir_base=savedir_base, columns=['train_loss', 'exp_id'])
    >>> print(lists_of_score_lists)
    """
    if len(exp_list) == 0:
        if verbose:
            print("exp_list is empty...")
        return

    exp_list = hu.filter_exp_list(exp_list, filterby_list, savedir_base=savedir_base, verbose=verbose)
    score_lists = []

    if return_as_dict:
        from collections import OrderedDict

        score_lists_dict = OrderedDict()
    # aggregate results
    for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
        savedir = os.path.join(savedir_base, exp_id)

        score_list_fname = os.path.join(savedir, score_list_name)
        if not os.path.exists(score_list_fname):
            if verbose:
                print("%s: missing %s" % (exp_id, score_list_name))
            continue

        else:
            s_list = hu.load_pkl(score_list_fname)
            if return_as_dict:
                score_lists_dict[exp_id] = s_list
            else:
                score_lists += [s_list]
    if return_as_dict:
        return score_lists_dict
    return score_lists


def get_exp_list_df(exp_list, filterby_list=None, columns=None, verbose=True):
    """Get a table showing the configurations for the given list of experiments

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    columns : list, optional
        a list of columns you would like to display, by default None

    Returns
    -------
    DataFrame
        a dataframe showing the scores obtained by the experiments

    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base,
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> df = hr.get_exp_list_df(exp_list, columns=['train_loss', 'exp_id'])
    >>> print(df)
    """
    if len(exp_list) == 0:
        if verbose:
            print("exp_list is empty...")
        return

    exp_list = hu.filter_exp_list(exp_list, filterby_list, verbose=verbose)

    # aggregate results
    result_list = []
    for exp_dict in exp_list:
        result_dict = {}

        exp_id = hu.hash_dict(exp_dict)
        result_dict["exp_id"] = exp_id

        for k in exp_dict:
            result_dict[k] = exp_dict[k]

        result_list += [result_dict]

    df = pd.DataFrame(result_list)

    if columns:
        df = df[[c for c in columns if c in df.columns]]

    return df



import tqdm
import argparse
import os
import gdown

from haven import haven_examples as he
from haven import haven_wizard as hw
from haven import haven_results as hr


if __name__ == "__main__":
    # Specify arguments regarding save directory and job scheduler
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--url",
        required=False,
        help="Define the experiment group to run.",
    )
    parser.add_argument("--path", required=False, help="Define the base directory where the experiments will be saved.")

    args, others = parser.parse_known_args()
    args.url = "https://drive.google.com/u/0/uc?id=1ByWQ49BzZiGrf_xlOjqps2RIT6hprjdi&export=download"
    args.path = "/mnt/public/datasets/matting/SIMD.tar"
    gdown.cached_download(args.url, path=args.path, quiet=False, proxy=False)



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu
from .. import haven_share as hd

import os, glob
import pprint
import json
import copy
import pprint
import pandas as pd
from . import widgets as wdg

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except Exception:
    print("widgets not available...")


def text_tab(db, output):
    blogs = widgets.Button(description="Get Predictions")

    button = widgets.VBox(
        [
            widgets.HBox([blogs]),
        ]
    )
    output_plot = widgets.Output()

    with output:
        display(button)
        display(output_plot)

    def on_logs_clicked(b):
        output_plot.clear_output()
        with output_plot:
            text_list = []
            for exp_dict in db.rm.exp_list:
                exp_id = hu.hash_dict(exp_dict)
                savedir_text = os.path.join(db.rm.savedir_base, exp_id, "text")
                if os.path.exists(savedir_text):
                    path = os.path.join(savedir_text, "*.json")
                    out = hu.load_json(glob.glob(path)[0])
                    text_list += [out]
                else:
                    text_list += ["None"]

            n_logs = len(text_list)

            for i, (exp_dict, text) in enumerate(zip(db.rm.exp_list, text_list)):
                exp_id = hu.hash_dict(exp_dict)
                savedir = os.path.join(db.rm_original.savedir_base, exp_id)
                print("\nPredictions %d/%d" % (i + 1, n_logs), "=" * 50)
                print("exp_id:", exp_id)
                print("savedir:", savedir)

                print("\nexp_dict")
                print("-" * 50)
                pprint.pprint(exp_dict)

                if text is not "None":
                    print("\nscore_dict")
                    print("-" * 50)
                    score_list = hu.load_pkl(os.path.join(savedir, "score_list.pkl"))
                    pprint.pprint(score_list[-1])

                print("\nPredictions")
                print("-" * 50)
                for j, t in enumerate(text):
                    print(f"\n*** Example {j} ***\n")
                    pprint.pprint(t)

    blogs.on_click(on_logs_clicked)



import matplotlib
import numpy as np
import os
import pandas as pd
import pylab as plt

from itertools import product
from matplotlib.ticker import ScalarFormatter, FuncFormatter
from sklearn.metrics.pairwise import pairwise_distances


matplotlib.style.use("bmh")

markers = [
    ("-", "o"),
    ("-", "p"),
    ("-", "D"),
    ("-", "^"),
    ("-", "s"),
    ("-", "8"),
    ("-", "o"),
    ("-", "o"),
    ("-", "o"),
    ("-", "o"),
    ("-", "o"),
    ("-", "o"),
]

colors = [
    "#741111",
    "#000000",
    "#3a49ba",
    "#7634c9",
    "#4C9950",
    "#CC29A3",
    "#ba3a3a",
    "#0f7265",
    "#7A7841",
    "#00C5CD",
    "#6e26d9",
]

bright_colors = ["#00C5CD"]


def myticks(x, pos):

    if x == 0:
        return "$0$"

    exponent = int(np.log10(x))
    coeff = x / 10 ** exponent

    # return r"{:0.1f}e{:0d}".format(coeff,exponent)

    return r"${:0.1f} \times 10^{{ {:2d} }}$".format(coeff, exponent)


def myticks_new(x, pos, exponent=1e5):

    if x == 0:
        return "$0$"

    exponent = int(np.log10(x))
    coeff = x / 10 ** exponent

    return r"${:0s}$".format(coeff / exponent)

    # return r"${:0.1f} \times 10^{{ {:2d} }}$".format(coeff,exponent)


class FixedOrderFormatter(ScalarFormatter):
    """Formats axis ticks using scientific notation with a constant order of
    magnitude"""

    def __init__(self, order_of_mag=0, useOffset=True, useMathText=False):
        self._order_of_mag = order_of_mag
        ScalarFormatter.__init__(self, useOffset=useOffset, useMathText=useMathText)

    def _set_orderOfMagnitude(self, range):
        """Over-riding this to avoid having orderOfMagnitude reset elsewhere"""
        self.orderOfMagnitude = self._order_of_mag


class PrettyPlot:
    def __init__(
        self,
        title=None,
        ylabel=None,
        xlabel=None,
        fontsize=14,
        linewidth=2.5,
        markersize=12,
        axFontSize=18,
        figsize=(13, 10),
        legend_type="line",
        yscale="log",
        subplots=(1, 1),
        shareRowLabel=True,
        axTickSize=14,
        legend_size=10,
        box_linewidth=1,
    ):
        self.box_linewidth = box_linewidth
        self.legend_size = legend_size
        self.axTickSize = int(axTickSize)
        self.fontsize = int(fontsize)
        self.shareRowLabel = shareRowLabel
        self.lim_set = False
        self.ylim = None
        self.legend_type = legend_type
        self.yscale = yscale
        self.linewidth = int(linewidth)
        self.markersize = int(markersize)
        self.axFontSize = int(axFontSize)

        if self.yscale == "log":
            plt.yscale("log")
        # ax.set_yscale('logit')
        self.labels = []
        self.y_list = []
        self.x_list = []
        self.converged = []
        fig = plt.figure(figsize=figsize)

        if title is not None:
            fig.suptitle(title, fontsize=self.axFontSize)
        self.fig = fig

        subplots = list(subplots)
        self.nrows = subplots[0]
        self.ncols = subplots[1]
        self.pIndex = 1

        self.axList = []

    def add_yxList(self, y_vals, x_vals, label, converged=False):
        if isinstance(y_vals, list):
            y_vals = np.array(y_vals)
        if isinstance(x_vals, list):
            x_vals = np.array(x_vals)

        self.y_list += [y_vals]
        self.x_list += [x_vals]

        self.labels += [label]

        self.converged += [converged]

    def show(self):
        plt.show()

    def save(self, path, iformat="png"):
        create_dirs(path)
        fname = path + ".%s" % iformat
        self.fig.savefig(fname, bbox_inches="tight")
        print(("Figure saved in %s" % (fname)))

    def plot_DataFrame(self, results):
        n_points, n_labels = results.shape

        x_vals = np.arange(n_points)
        labels = results.columns
        y_array = np.array(results)
        y_list = []
        x_list = []
        for j in range(n_labels):
            x_list += [x_vals]
            y_list += [y_array[:, j]]

        self.plot(y_list, x_list, labels)

    def set_lim(self, ylim, xlim):
        self.lim_set = True
        self.ylim = ylim

        self.ax.set_ylim(ylim)
        self.ax.set_xlim(xlim)

    def set_tickSize(self, labelsize=8):
        [tick.label.set_fontsize(labelsize) for tick in self.ax.yaxis.get_major_ticks()]
        [tick.label.set_fontsize(labelsize) for tick in self.ax.xaxis.get_major_ticks()]

    def set_title(self, title):
        self.fig.suptitle(title, fontsize=self.axFontSize, y=1.08)

    def plot(self, y_list=None, x_list=None, labels=None, ax=None, ylabel="", xlabel="", yscale=False):
        fig = self.fig

        if y_list == None and x_list == None:
            y_list = self.y_list
            x_list = self.x_list

        if yscale == "log":
            # Makse sure everything is non-negative
            # for yi in y_list:
            #     assert np.all(yi >= 0)

            # Set zeros to eps
            for i in range(len(y_list)):
                y_list[i] = np.maximum(y_list[i], np.finfo(float).eps)

            # Set zeros to eps
            for i in range(len(y_list)):

                opt_ind = np.where(y_list[i] == np.finfo(float).eps)[0]
                if opt_ind.size > 0:
                    opt_ind = opt_ind[0]

                    y_list[i] = y_list[i][: opt_ind + 1]
                    x_list[i] = x_list[i][: opt_ind + 1]

        n_labels = len(y_list)

        if ax is None:
            ax = self.fig.add_subplot(self.nrows, self.ncols, self.pIndex)

        ax.set_facecolor("white")
        ax.set_yscale("log", nonposy="clip")
        if labels is None and self.labels is None:
            labels = list(map(str, np.arange(n_labels)))
        elif labels is None:
            labels = self.labels

        ref_points = []
        for i in range(len(self.converged)):
            if self.converged[i] is not None:

                ref_points += [[self.converged[i]["X"], self.converged[i]["Y"]]]

        label_positions, label_indices = get_labelPositions(
            y_list, x_list, self.ylim, labels=labels, ref_points=np.array(ref_points)
        )

        ls_markers = markers

        if not self.lim_set:
            y_min, y_max = get_min_max(y_list)
            x_min, x_max = get_min_max(x_list)
            # y_min = max(y_min, 1e-8)
            ax.set_ylim([y_min, y_max])
            ax.set_xlim([x_min, x_max])

        for i in range(n_labels):
            color = colors[i]
            ls, marker = ls_markers[i]

            y_vals = y_list[i]
            x_vals = x_list[i]

            n_points = len(y_vals)

            label = labels[i]

            markerFreq = n_points / (int(np.log(n_points)) + 1)

            # SCATTER PLOT OPTIMAL
            # ind_opt = np.where(y_vals == np.finfo(float).eps)[0]

            # if ind_opt.size > 0:
            #     x_opt = x_vals[np.where(y_vals == np.finfo(float).eps)[0][0]]
            #     y_opt = np.finfo(float).eps

            if self.converged[i] is not None:
                ax.scatter(
                    self.converged[i]["X"],
                    self.converged[i]["Y"],
                    s=300,
                    marker="*",
                    color=color,
                    clip_on=False,
                    zorder=100,
                )
            ##
            (line,) = ax.plot(
                x_vals,
                y_vals,
                markevery=int(markerFreq),
                markersize=int(self.markersize),
                color=color,
                lw=self.linewidth,
                alpha=1.0,
                label=label,
                ls=ls,
                marker=marker,
            )

            if self.legend_type == "line":
                x_point, y_point = label_positions[i]
                angle = get_label_angle(x_vals, y_vals, label_indices[i], ax, color="0.5", size=12)

                box = dict(
                    facecolor="white",
                    edgecolor=color,
                    linestyle=ls,
                    # hatch=marker,
                    linewidth=int(2),
                    boxstyle="round",
                )

                ax.text(
                    x_point,
                    y_point,
                    label,
                    va="center",
                    ha="center",
                    rotation=angle,
                    color=color,
                    bbox=box,
                    fontsize=self.legend_size,
                )

            else:
                plt.legend(loc="best")

        if self.shareRowLabel and (((self.pIndex - 1) % (self.ncols)) == 0):
            ax.set_ylabel(ylabel, fontsize=self.axFontSize)

        if not self.shareRowLabel:
            ax.set_ylabel(ylabel, fontsize=self.axFontSize * 1.1)

        ax.set_xlabel(xlabel, fontsize=self.axFontSize * 1.1)

        ax.tick_params(labelsize=self.axTickSize * 1.3)
        ax.tick_params(axis="y", labelsize=int(self.axTickSize * 1.5))
        self.y_list = []
        self.x_list = []
        self.labels = []
        self.converged = []

        self.pIndex += 1
        self.axList += [ax]

        ax.minorticks_off()
        vals = np.logspace(np.log10(y_min), np.log10(y_max), 5)
        ax.set_yticks(vals)

        ax.yaxis.set_major_formatter(FuncFormatter(myticks))

        return fig, ax


#########
# HELPERS
#########
def get_overlapPercentage(index, y_list):
    n_points = y_list[0].size
    for i in range(index + 1):
        n_points = min(n_points, y_list[i].size)

    y_vector = y_list[index][:n_points, np.newaxis]

    prev_lines = np.zeros((n_points, index))

    for i in range(index):
        prev_lines[:, i] = y_list[i][:n_points]

        prev_lines /= np.linalg.norm(prev_lines, axis=0) + 1e-10

    y_norm = y_vector / np.linalg.norm(y_vector, axis=0)

    diff = np.abs((prev_lines - y_norm)).min(axis=1)

    n_overlap = np.sum(diff < 1e-6)

    percentage = n_overlap / float(n_points)

    return percentage


def create_dirs(fname):
    if "/" not in fname:
        return

    if not os.path.exists(os.path.dirname(fname)):
        try:
            os.makedirs(os.path.dirname(fname))
        except OSError:
            pass


def normalize(xy_points, ref_points, y_min, y_max, x_min, x_max):

    xy_points[:, 1] = np.log(np.maximum(1e-15, xy_points[:, 1])) / np.log(10)
    ref_points[:, 1] = np.log(np.maximum(ref_points[:, 1], 1e-15)) / np.log(10)

    y_min = np.log(y_min) / np.log(10)
    y_max = np.log(y_max) / np.log(10)

    xy_normed = xy_points - np.array([x_min, y_min])
    xy_normed /= np.array([x_max - x_min, y_max - y_min])

    ref_normed = ref_points - np.array([x_min, y_min])

    ref_normed /= np.array([x_max - x_min, y_max - y_min])
    return xy_normed, ref_normed


# LABEL POSITIONS


def get_labelPositions(y_list, x_list, ylim=None, labels=None, ref_points=None):
    if ref_points is None:

        ref_points = []
    """Get label positions greedily"""

    n_labels = len(y_list)

    # GET BORDER POINTS
    x_min, x_max = get_min_max(x_list)
    if ylim is not None:
        y_min, y_max = ylim
    else:
        y_min, y_max = get_min_max(y_list)

    xdiff = x_max - x_min
    ydiff = y_max - y_min

    # Border points
    bp1 = np.array(list(product([x_min, x_max, xdiff * 0.5], [y_min, y_max, ydiff * 0.5])))[:-1]
    bp1 = np.array(list(product([x_max], [y_max])))[:-1]

    bp1 = np.array(list(product([8], [0])))

    addedPoints = []

    for yPoint in np.linspace(y_min, y_max, 6):
        addedPoints += [(x_min, yPoint)]
        addedPoints += [(x_max, yPoint)]

    sPoints = [(xx[0], yy[0]) for xx, yy in zip(x_list, y_list)]
    ePoints = [(xx[-1], yy[-1]) for xx, yy in zip(x_list, y_list)]
    bp2 = np.array(addedPoints + sPoints + ePoints)
    if len(ref_points) == 0:
        border_points = np.vstack([bp1, bp2])
    else:
        border_points = np.vstack([bp1, bp2, ref_points])

    n_border = border_points.shape[0]

    # Initialize placeholders
    ref_points = np.zeros((n_border + n_labels, 2))

    label_positions = np.zeros((n_labels, 2))
    label_indices = np.zeros(n_labels, int)

    ref_points[:n_border] = border_points

    for i in range(n_labels):
        # GET POSITIONS

        if ylim is not None:
            ind = (y_list[i] < y_max + 1e-4) & (y_list[i] > y_min - 1e-4)
            n_points = x_list[i][ind].size
            xy_points = np.zeros((n_points, 2))
            xy_points[:, 0] = x_list[i][ind]
            xy_points[:, 1] = y_list[i][ind]
        else:
            n_points = x_list[i].size
            xy_points = np.zeros((n_points, 2))
            xy_points[:, 0] = x_list[i]
            xy_points[:, 1] = y_list[i]

        # NORMALIZE

        xy_normed, ref_normed = normalize(
            xy_points.copy(), ref_points[: n_border + i].copy(), y_min, y_max, x_min, x_max
        )
        # GET REF POINTS

        dist = pairwise_distances(xy_normed, ref_normed, metric="l1")

        # GET MINIMUM DISTANCES
        min_dist = dist.min(axis=1)

        # GET MAXIMUM MINIMUM DISTANCE
        label_index = np.argmax(min_dist)
        label_pos = xy_points[label_index]

        ref_points[n_border + i] = label_pos
        label_positions[i] = label_pos
        label_indices[i] = label_index

    return label_positions, label_indices


def get_min_max(v_list):
    vector = v_list[0]
    v_min = np.min(vector)
    v_max = np.max(vector)

    for i in range(1, len(v_list)):
        vector = v_list[i]
        v_min = min(np.min(vector), v_min)
        v_max = max(np.max(vector), v_max)

    return v_min, v_max


def get_label_angle(xdata, ydata, index, ax, color="0.5", size=12, window=3):
    n_points = xdata.size

    x1 = xdata[index]
    y1 = ydata[index]

    # ax = line.get_axes()

    sp1 = ax.transData.transform_point((x1, y1))

    slope_degrees = 0.0
    count = 0.0

    for i in range(index + 1, min(index + window, n_points)):
        y2 = ydata[i]
        x2 = xdata[i]

        sp2 = ax.transData.transform_point((x2, y2))

        rise = sp2[1] - sp1[1]
        run = sp2[0] - sp1[0]

        slope_degrees += np.degrees(np.arctan2(rise, run))
        count += 1.0

    for i in range(index - 1, max(index - window, 0), -1):
        y2 = ydata[i]
        x2 = xdata[i]

        sp2 = ax.transData.transform_point((x2, y2))

        rise = -(sp2[1] - sp1[1])
        run = -(sp2[0] - sp1[0])

        slope_degrees += np.degrees(np.arctan2(rise, run))
        count += 1.0

    slope_degrees /= count

    return slope_degrees


def box_color(edgecolor, linestyle, marker):
    """Creates box shape"""
    return dict(
        facecolor="white",
        edgecolor=edgecolor,
        linestyle=linestyle,
        # hatch=marker,
        linewidth=2,
        boxstyle="round",
    )



import argparse
import pandas as pd


def joke():
    sdfds
    parser = argparse.ArgumentParser()

    parser.add_argument("-e", "--exp_group_list")
    parser.add_argument("-sb", "--savedir_base", required=True)
    parser.add_argument("-ei", "--exp_id", default=None)

    args = parser.parse_args()
    print(args)


def run_file():
    print("lol")



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu
from .. import haven_share as hd
from . import widgets as wdg

import os
import pprint
import json
import copy
import pprint
import pandas as pd

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except Exception:
    print("widgets not available...")


def plots_tab(self, output):
    db = self

    llegend_format = widgets.Text(value=str(self.vars.get("legend_format", "")), description="", disabled=False)
    ltitle_format = widgets.Text(value=str(self.vars.get("title_format", "")), description="", disabled=False)

    lcmap = widgets.Text(
        value=str(self.vars.get("cmap", "jet")), description="cmap:", layout=self.layout_dropdown, disabled=False
    )

    llog_metric_list = widgets.Text(
        value=str(self.vars.get("log_metric_list", "[train_loss]")), description="log_metric_list:", disabled=False
    )

    bdownload = widgets.Button(description="Download Plots")
    bdownload_out = widgets.Output(layout=self.layout_button)

    def on_download_clicked(b):
        fname = "plots.pdf"
        from matplotlib.backends.backend_pdf import PdfPages
        import matplotlib.pyplot as plt

        pp = PdfPages(fname)
        for fig in self.rm_original.fig_list:
            fig.savefig(pp, format="pdf")
        pp.close()

        bdownload_out.clear_output()

        with bdownload_out:
            display(FileLink(fname, result_html_prefix="Download: "))

    bdownload.on_click(on_download_clicked)

    h22 = widgets.Label(
        value="Format:",
        layout=widgets.Layout(width="340px"),
    )

    h33 = widgets.Label(
        value="Format:",
        layout=widgets.Layout(width="340px"),
    )

    h44 = widgets.Label(
        value="",
        layout=widgets.Layout(width="340px"),
    )

    space = widgets.Label(
        value="",
        layout=widgets.Layout(width="300px"),
    )
    brefresh = widgets.Button(description="Display Plot")
    d_avg_across_txt = widgets.Label(
        value="avg_across:",
    )

    w_y_metrics = wdg.SelectMultiple(
        header="Y-axis Metrics:", options=self.rm_original.score_keys, db_vars=db.vars, var="y_metrics"
    )

    if db.vars.get("legend_list") is None:
        db.vars["legend_list"] = hu.get_diff_hparam(db.rm.exp_list)
    w_legend = wdg.SelectMultiple(
        header="Legend:", options=["exp_id"] + db.rm.exp_params, db_vars=db.vars, var="legend_list", select_all=True
    )

    w_title = wdg.SelectMultiple(header="Title:", options=db.rm.exp_params, db_vars=db.vars, var="title_list")

    w_groupby = wdg.SelectMultiple(
        header="GroupBy:", options=["None"] + db.rm.exp_params, db_vars=db.vars, var="groupby_list"
    )

    w_x_metric = wdg.Dropdown(
        header="X-axis Metric", options=self.rm_original.score_keys, db_vars=db.vars, var="x_metric"
    )
    w_mode = wdg.Dropdown(header="Plot Mode", options=["line", "bar"], db_vars=db.vars, var="mode")
    w_bar_agg = wdg.Dropdown(
        header="Plot Agg (bar plot only) ", options=["last", "max", "mean", "min"], db_vars=db.vars, var="bar_agg"
    )
    w_avg_across = wdg.Dropdown(
        header="Avg Across", options=["None"] + db.rm.exp_params, db_vars=db.vars, var="avg_across"
    )

    button = widgets.VBox(
        [
            widgets.HBox(
                [
                    w_y_metrics.get_widget(),
                    w_legend.get_widget(),
                    w_title.get_widget(),
                    w_groupby.get_widget(),
                ]
            ),
            widgets.HBox(
                [
                    w_x_metric.get_widget(),
                    w_mode.get_widget(),
                    w_bar_agg.get_widget(),
                    w_avg_across.get_widget(),
                ]
            ),
            #    widgets.HBox([ d_avg_across_txt, d_avg_across_columns,  ]),
            widgets.HBox(
                [
                    brefresh,
                    bdownload,
                    bdownload_out,
                ]
            ),
        ]
    )

    output_plot = widgets.Output()

    def on_clicked(b):
        output_plot.clear_output()
        with output_plot:
            self.update_rm()

            w, h = 10, 5
            if len(w_y_metrics.update()) == 0:
                display("No results saved yet.")
                return

            elif len(w_y_metrics.update()) > 1:
                figsize = (2 * int(w), int(h))
                self.vars["figsize"] = figsize
            else:
                self.vars["figsize"] = (int(w), int(h))

            self.vars["legend_format"] = llegend_format.value
            self.vars["log_metric_list"] = hu.get_list_from_str(llog_metric_list.value)

            self.vars["title_format"] = ltitle_format.value
            self.vars["cmap"] = lcmap.value

            self.rm_original.fig_list = self.rm.get_plot_all(
                y_metric_list=w_y_metrics.update(),
                x_metric=w_x_metric.update(),
                groupby_list=w_groupby.update(),
                legend_list=w_legend.update(),
                log_metric_list=self.vars["log_metric_list"],
                mode=w_mode.update(),
                bar_agg=w_bar_agg.update(),
                figsize=self.vars["figsize"],
                title_list=w_title.update(),
                legend_format=self.vars["legend_format"],
                title_format=self.vars["title_format"],
                cmap=self.vars["cmap"],
                avg_across=w_avg_across.update(),
            )

            show_inline_matplotlib_plots()

    brefresh.on_click(on_clicked)

    with output:
        display(button)
        display(output_plot)



import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
import pylab as plt
from skimage.segmentation import mark_boundaries

try:
    import cv2
except:
    # ignore cv2 if it can't get imported, sometimes it doesn't install well
    pass

from . import haven_utils as hu


def mask_on_image(image, mask, add_bbox=False, return_pil=False):
    """[summary]

    Parameters
    ----------
    image : [type]
        [description]
    mask : [type]
        [description]
    add_bbox : bool, optional
        [description], by default True

    Returns
    -------
    [type]
        [description]
    """
    image = image_as_uint8(image)
    mask = np.array(mask).squeeze()
    obj_ids = np.unique(mask)

    # polygons = cv2.findContours(im,cv2.RETR_TREE,cv2.CHAIN_APPROX_NONE)[1][0]
    red = np.zeros(image.shape, dtype="uint8")
    red[:, :, 2] = 255
    alpha = 0.5
    result = image.copy()
    for o in obj_ids:
        if o == 0:
            continue
        ind = mask == o
        result[ind] = result[ind] * alpha + red[ind] * (1 - alpha)
        pos = np.where(ind)
        xmin = np.min(pos[1])
        xmax = np.max(pos[1])
        ymin = np.min(pos[0])
        ymax = np.max(pos[0])
        if add_bbox:
            result = cv2.rectangle(result, (xmin, ymin), (xmax, ymax), color=(0, 255, 0), thickness=2)
    result = mark_boundaries(result, mask)

    if return_pil:
        return Image.fromarray(result)

    return result


def resize_points(points, h, w):
    """[summary]

    Parameters
    ----------
    points : [type]
        [description]
    h : [type]
        [description]
    w : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    points = points.squeeze()
    h_old, w_old = points.shape
    y_list, x_list = np.where(points.squeeze())

    points_new = np.zeros((h, w))

    for y, x in zip(y_list, x_list):
        y_new = int((y / h_old) * h)
        x_new = int((x / w_old) * w)
        points_new[y_new, x_new] = 1

    return points_new


def gray2cmap(gray, cmap="jet", thresh=0):
    """gets a heatmap for a given gray image. Can be used to visualize probabilities.

    Parameters
    ----------
    gray : [type]
        [description]
    cmap : str, optional
        [description], by default "jet"
    thresh : int, optional
        [description], by default 0

    Returns
    -------
    [type]
        [description]
    """
    # Gray has values between 0 and 255 or 0 and 1
    gray = hu.t2n(gray)
    gray = gray / max(1, gray.max())
    gray = np.maximum(gray - thresh, 0)
    gray = gray / max(1, gray.max())
    gray = gray * 255

    gray = gray.astype(int)
    # print(gray)

    from matplotlib.cm import get_cmap

    cmap = get_cmap(cmap)

    output = np.zeros(gray.shape + (3,), dtype=np.float64)

    for c in np.unique(gray):
        output[(gray == c).nonzero()] = cmap(c)[:3]

    return hu.l2f(output)


def scatter_plot(X, color, fig=None, title=""):
    """[summary]

    Parameters
    ----------
    X : [type]
        [description]
    color : [type]
        [description]
    fig : [type], optional
        [description], by default None
    title : str, optional
        [description], by default ""

    Returns
    -------
    [type]
        [description]
    """
    if fig is None:
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_subplot(1, 1, 1)

    ax = fig.axes[0]

    ax.grid(linestyle="dotted")
    ax.scatter(X[:, 0], X[:, 1], alpha=0.6, c=color, edgecolors="black")

    # plt.axes().set_aspect('equal', 'datalim')
    ax.set_title(title)
    ax.set_xlabel("t-SNE Feature 2")
    ax.set_ylabel("t-SNE Feature 1")

    fig.tight_layout(rect=[0, 0.03, 1, 0.95])

    return fig


def pretty_vis(image, annList, show_class=False, alpha=0.0, dpi=100, **options):
    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.patches import Polygon
    from matplotlib.figure import Figure
    from . import ann_utils as au

    # print(image)
    # if not image.as > 1:
    #     image = image.astype(float)/255.
    image = f2l(image).squeeze().clip(0, 255)
    if image.max() > 1:
        image /= 255.0

    # box_alpha = 0.5
    # print(image.clip(0, 255).max())
    color_list = colormap(rgb=True) / 255.0

    # fig = Figure()
    fig = plt.figure(frameon=False)
    canvas = FigureCanvas(fig)
    fig.set_size_inches(image.shape[1] / dpi, image.shape[0] / dpi)
    # ax = fig.gca()

    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
    ax.axis("off")
    fig.add_axes(ax)
    # im = im.clip(0, 1)
    # print(image)
    ax.imshow(image)

    # Display in largest to smallest order to reduce occlusion
    # areas = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])
    # sorted_inds = np.argsort(-areas)

    mask_color_id = 0
    for i in range(len(annList)):
        ann = annList[i]

        # bbox = boxes[i, :4]
        # score = boxes[i, -1]

        # bbox = au.ann2bbox(ann)["shape"]
        # score = ann["score"]

        # if score < thresh:
        #     continue

        # show box (off by default, box_alpha=0.0)
        if "bbox" in ann:
            bbox = ann["bbox"]
            ax.add_patch(
                plt.Rectangle((bbox[0], bbox[1]), bbox[2], bbox[3], fill=False, edgecolor="r", linewidth=3.0, alpha=0.5)
            )

        # if show_class:
        # if options.get("show_text") == True or options.get("show_text") is None:
        #     score = ann["score"] or -1
        #     ax.text(
        #         bbox[0], bbox[1] - 2,
        #         "%.1f" % score,
        #         fontsize=14,
        #         family='serif',
        #         bbox=dict(facecolor='g', alpha=1.0, pad=0, edgecolor='none'),
        #         color='white')

        # show mask
        if "segmentation" in ann:
            mask = au.ann2mask(ann)["mask"]
            img = np.ones(image.shape)
            # category_id = ann["category_id"]
            # mask_color_id = category_id - 1
            # color_list = ["r", "g", "b","y", "w","orange","purple"]
            # color_mask = color_list[mask_color_id % len(color_list)]
            color_mask = color_list[mask_color_id % len(color_list), 0:3]
            mask_color_id += 1
            # print("color id: %d - category_id: %d - color mask: %s"
            # %(mask_color_id, category_id, str(color_mask)))
            w_ratio = 0.4
            for c in range(3):
                color_mask[c] = color_mask[c] * (1 - w_ratio) + w_ratio
            for c in range(3):
                img[:, :, c] = color_mask[c]
            e = mask

            contour, hier = cv2.findContours(e.copy(), cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE)

            for c in contour:
                polygon = Polygon(
                    c.reshape((-1, 2)), fill=True, facecolor=color_mask, edgecolor="white", linewidth=1.5, alpha=0.7
                )
                ax.add_patch(polygon)

    canvas.draw()  # draw the canvas, cache the renderer
    width, height = fig.get_size_inches() * fig.get_dpi()
    # image = np.fromstring(canvas.tostring_rgb(), dtype='uint8')

    fig_image = np.fromstring(canvas.tostring_rgb(), dtype="uint8").reshape(int(height), int(width), 3)
    plt.close()
    # print(fig_image)
    return fig_image


def text_on_image(text, image):
    """Adds test on the image

    Parameters
    ----------
    text : [type]
        [description]
    image : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    font = cv2.FONT_HERSHEY_SIMPLEX
    bottomLeftCornerOfText = (10, 40)
    fontScale = 0.8
    fontColor = (1, 1, 1)
    lineType = 1
    # img_mask = skimage.transform.rescale(np.array(img_mask), 1.0)
    # img_np = skimage.transform.rescale(np.array(img_points), 1.0)
    img_np = cv2.putText(
        image,
        text,
        bottomLeftCornerOfText,
        font,
        fontScale,
        fontColor,
        thickness=2
        # lineType
    )
    return img_np


def bbox_on_image(bbox_xyxy, image, mode="yxyx", color=(255, 0, 0)):
    """[summary]

    Parameters
    ----------
    bbox_xyxy : [type]
        [description]
    image : [type]
        [description]
    mode : str, optional
        [description], by default 'xyxy'

    Returns
    -------
    [type]
        [description]
    """
    import cv2

    image_uint8 = image_as_uint8(image)

    H, W, _ = image_uint8.shape

    for bb in bbox_xyxy:
        if mode == "yxyx":
            y1, x1, y2, x2 = bb
        else:
            x1, y1, x2, y2 = bb

        if mode == "xywh":
            x2 += x1
            y2 += y1

        if x2 < 1:
            start_point = (
                int(x1 * W),
                int(y1 * H),
            )
            end_point = (
                int(x2 * W),
                int(y2 * H),
            )
        else:
            start_point = (int(x1), int(y1))
            end_point = (int(x2), int(y2))

        # Line thickness of 2 px
        thickness = 2
        # Draw a rectangle with blue line borders of thickness of 2 px
        image_uint8 = cv2.rectangle(image_uint8.copy(), start_point, end_point, color, thickness)

    return image_uint8 / 255.0


def points_on_image(y_list, x_list, image, radius=3, c_list=None):
    """[summary]

    Parameters
    ----------
    y_list : [type]
        [description]
    x_list : [type]
        [description]
    image : [type]
        [description]
    radius : int, optional
        [description], by default 3

    Returns
    -------
    [type]
        [description]
    """
    image_uint8 = image_as_uint8(image)

    H, W, _ = image_uint8.shape
    color_list = [(255, 0, 0), (0, 255, 0), (0, 0, 255)]
    for i, (y, x) in enumerate(zip(y_list, x_list)):
        if y < 1:
            x, y = int(x * W), int(y * H)
        else:
            x, y = int(x), int(y)

        # Blue color in BGR
        if c_list is not None:
            color = color_list[c_list[i]]
        else:
            color = color_list[1]

        # Line thickness of 2 px
        thickness = 5
        # Using cv2.rectangle() method
        # Draw a rectangle with blue line borders of thickness of 2 px
        image_uint8 = cv2.circle(image_uint8, (x, y), radius, color, thickness)

        start_point = (x - radius * 2, y - radius * 2)
        end_point = (x + radius * 2, y + radius * 2)
        thickness = 2
        color = (255, 0, 0)

        image_uint8 = cv2.rectangle(image_uint8, start_point, end_point, color, thickness)

    return image_uint8 / 255.0


def image_as_uint8(img):
    """Returns a uint8 version of the image

    Parameters
    ----------
    img : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    image = hu.f2l(np.array(img).squeeze())

    if image.dtype != "uint8":
        image_uint8 = (image * 255).astype("uint8").copy()
    else:
        image_uint8 = image

    return image_uint8



from haven import haven_wizard as hw
import wandb
import sys
import os
import pprint

path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0, path)


if __name__ == "__main__":
    # first way
    score_dict = {"loss": loss}
    wandb.send(score_dict)

    # second way
    chk = load_checkpoint(savedir)
    hw.save_checkpoint(savedir, score_dict=score_dict, wandb_config={})



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu
from .. import haven_share as hd

from .share_tab import share_tab
from .plots_tab import plots_tab
from .tables_tab import tables_tab
from .latex_tab import latex_tab
from .images_tab import images_tab
from .text_tab import text_tab
from . import widgets as wdg

import haven
import os
import pprint
import json
import copy
import pprint
import pandas as pd

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except Exception:
    print("widgets not available...")


def get_dashboard(rm, vars=None, show_jobs=True, wide_display=False, enable_datatables=True):
    dm = DashboardManager(
        rm, vars=vars, show_jobs=show_jobs, wide_display=wide_display, enable_datatables=enable_datatables
    )
    dm.display()
    return dm


class DashboardManager:
    def __init__(self, rm, vars=None, show_jobs=True, wide_display=True, enable_datatables=True):
        self.rm_original = rm
        if vars is None:
            self.vars = {}
        else:
            self.vars = vars

        self.show_jobs = show_jobs
        self.wide_display = wide_display
        self.enable_datatables = enable_datatables

        self.layout = widgets.Layout(width="100px")
        self.layout_label = widgets.Layout(width="200px")
        self.layout_dropdown = widgets.Layout(width="200px")
        self.layout_button = widgets.Layout(width="200px")
        self.t_savedir_base = widgets.Text(
            value=str(self.vars.get("savedir_base") or rm.savedir_base),
            layout=widgets.Layout(width="600px"),
            disabled=False,
        )

        self.t_filterby_list = widgets.Text(
            value=str(self.vars.get("filterby_list")),
            layout=widgets.Layout(width="1200px"),
            description="               filterby_list:",
            disabled=False,
        )

    def display(self):
        self.update_rm(display_meta=False)

        header = widgets.Label(
            value="Loading Dashboard...",
            layout=widgets.Layout(width="800px"),
        )
        display(header)

        if self.enable_datatables:
            init_datatable_mode()
        tables = widgets.Output()
        plots = widgets.Output()
        images = widgets.Output()
        text = widgets.Output()
        share = widgets.Output()
        latex = widgets.Output()

        main_out = widgets.Output()
        # Display tabs
        tab = widgets.Tab(children=[tables, plots, images, text, latex, share])
        tab.set_title(1, "Plots")
        tab.set_title(0, "Tables")
        tab.set_title(2, "Images")
        tab.set_title(3, "Text")
        tab.set_title(4, "Latex")
        tab.set_title(5, "Share")

        with main_out:
            display(tab)
            tables.clear_output()
            plots.clear_output()
            images.clear_output()
            text.clear_output()
            latex.clear_output()
            share.clear_output()

            # show tabs
            tables_tab(self, tables)
            plots_tab(self, plots)
            images_tab(self, images)
            text_tab(self, text)
            latex_tab(self, latex)
            share_tab(self, share)

            header.value = (
                f"Dashboard loaded (ver: {haven.__version__}). "
                + f'{len(self.rm_original.exp_list_all)} experiments selected from "{self.rm_original.savedir_base}"'
            )

        display(main_out)

        if self.wide_display:
            display(HTML("<style>.container { width:100% !important; }</style>"))

        # This makes cell show full height display
        style = """
        <style>
            .output_scroll {
                height: unset !important;
                border-radius: unset !important;
                -webkit-box-shadow: unset !important;
                box-shadow: unset !important;
            }
        </style>
        """
        display(HTML(style))

    def update_rm(self, display_meta=True):
        self.rm = hr.ResultManager(
            exp_list=self.rm_original.exp_list_all,
            savedir_base=str(self.t_savedir_base.value),
            filterby_list=hu.get_dict_from_str(str(self.t_filterby_list.value)),
            verbose=self.rm_original.verbose,
            mode_key=self.rm_original.mode_key,
            has_score_list=self.rm_original.has_score_list,
            score_list_name=self.rm_original.score_list_name,
            job_scheduler=self.rm_original.job_scheduler,
        )

        # if len(self.rm.exp_list) == 0:
        #     if self.rm.n_exp_all > 0:
        #         display('No experiments selected out of %d '
        #                 'for filtrby_list %s' % (self.rm.n_exp_all,
        #                                          self.rm.filterby_list))
        #         display('Table below shows all experiments.')
        #         score_table = hr.get_score_df(exp_list=self.rm_original.exp_list_all,
        #                                       savedir_base=self.rm_original.savedir_base)
        #         display(score_table)
        #     else:
        #         display('No experiments exist...')
        #     return
        # else:
        if display_meta:
            display(f'{len(self.rm.exp_list)}/{len(self.rm.exp_list_all)} experiments selected using "filterby_list"')


def launch_jupyter():
    """
    virtualenv -p python3 .
    source bin/activate
    pip install jupyter notebook
    jupyter notebook --ip 0.0.0.0 --port 2222 --NotebookApp.token='abcdefg'
    """
    print()


def create_jupyter_file(
    fname,
    savedir_base,
    overwrite=False,
    print_url=False,
):
    savedir_base = os.path.abspath(savedir_base)
    if overwrite or not os.path.exists(fname):
        cells = [main_cell(savedir_base), sub_cell(), install_cell()]
        if os.path.dirname(fname) != "":
            os.makedirs(os.path.dirname(fname), exist_ok=True)
        save_ipynb(fname, cells)
        print("> Open %s to visualize results" % fname)

    if print_url:
        from notebook import notebookapp

        servers = list(notebookapp.list_running_servers())
        hostname = os.uname().nodename

        flag = False
        for i, s in enumerate(servers):
            if s["hostname"] == "localhost":
                continue
            flag = True
            url = "http://%s:%s/" % (hostname, s["port"])
            print("- url:", url)

        if flag == False:
            print("a jupyter server was not found :(")


def sub_cell():
    script = """
# get table 
rm.get_score_df().head()

# get latex 
# print(rm.get_latex_table(legend=['dataset'], metrics=['train_loss'], decimals=1, caption="Results", label='tab:results'))

# get custom plots
fig = rm.get_plot_all(
                # order='metrics_by_groups',
                # avg_across='runs',
                y_metric_list=y_metrics, 
                x_metric=x_metric,
                # legend_fontsize=18,
                # x_fontsize=20,
                # y_fontsize=20,
                # xtick_fontsize=20,
                # ytick_fontsize=20,
                # title_fontsize=24,
                # legend_list=['model], 
                # title_list = ['dataset'], 
                # title_format='Dataset:{}',
                # log_metric_list = ['train_loss'], 
                # groupby_list = ['dataset'],
                # map_ylabel_list=[{'train_loss':'Train loss'}],
                # map_xlabel_list=[{'epoch':'Epoch'}],
                # figsize=(15,5),
                # plot_confidence=False,
                # savedir_plots='%s' % (name)
)
          """
    return script


def main_cell(savedir_base):
    script = (
        """
from haven import haven_jupyter as hj
from haven import haven_results as hr
from haven import haven_utils as hu

# path to where the experiments got saved
savedir_base = '%s'
exp_list = None

# get experiments from the exp config
exp_config_fname = None
if exp_config_fname:
    config = hu.load_py(exp_config_fname)
    exp_list = []
    for exp_group in [
        "example"
                    ]:
        exp_list += config.EXP_GROUPS[exp_group]

# filter exps

filterby_list = None
# filterby_list =[{'dataset':'mnist'}]

# get experiments
rm = hr.ResultManager(exp_list=exp_list,
                      savedir_base=savedir_base,
                      filterby_list=filterby_list,
                      verbose=0,
                      exp_groups=None,
                      job_scheduler='slurm'
                     )

# specify display parameters

# groupby_list = ['dataset']
# title_list = ['dataset']
# legend_list = ['model']
y_metrics = ['train_loss']
x_metric = 'epoch'

# launch dashboard
hj.get_dashboard(rm, vars(), wide_display=False, enable_datatables=False)
          """
        % savedir_base
    )
    return script


def install_cell():
    script = """
    !pip install --upgrade git+https://github.com/haven-ai/haven-ai
          """
    return script


def save_ipynb(fname, script_list):
    import nbformat as nbf

    nb = nbf.v4.new_notebook()
    nb["cells"] = [nbf.v4.new_code_cell(code) for code in script_list]
    with open(fname, "w") as f:
        nbf.write(nb, f)


def init_datatable_mode():
    """Initialize DataTable mode for pandas DataFrame represenation."""
    import pandas as pd
    from IPython.core.display import display, Javascript

    # configure path to the datatables library using requireJS
    # that way the library will become globally available
    display(
        Javascript(
            """
        require.config({
            paths: {
                DT: '//cdn.datatables.net/1.10.19/js/jquery.dataTables.min',
            }
        });
        $('head').append('<link rel="stylesheet" type="text/css" href="//cdn.datatables.net/1.10.19/css/jquery.dataTables.min.css">');
    """
        )
    )

    def _repr_datatable_(self):
        """Return DataTable representation of pandas DataFrame."""
        # classes for dataframe table (optional)
        classes = ["table", "table-striped", "table-bordered"]

        # create table DOM
        script = f"$(element).html(`{self.to_html(index=True, classes=classes)}`);\n"

        # execute jQuery to turn table into DataTable
        script += """
            require(["DT"], function(DT) {
                $(document).ready( () => {
                    // Turn existing table into datatable
                    $(element).find("table.dataframe").DataTable({"scrollX": true});

                    $('#container').css( 'display', 'block' );
                    table.columns.adjust().draw();

                })
            });
        """

        return script

    pd.DataFrame._repr_javascript_ = _repr_datatable_



import os
import time
import sys
import subprocess
from .. import haven_utils as hu
from .. import haven_chk as hc
from textwrap import wrap

import copy
import pandas as pd
import numpy as np
import getpass
import pprint

ALIVE_STATES = ["RUNNING", "QUEUED", "PENDING", "QUEUING"]
COMPLETED_STATES = ["COMPLETED", "SUCCEEDED", "COMPLETING"]
FAILED_STATES = ["FAILED", "CANCELLED", "INTERRUPTED", "TIMEOUT", "NODE_FAIL"]
SUFFIX_TO_ADD_TO_SAVEDIR_TO_GET_WORKDIR = "code"


class JobManager:
    """Job manager."""

    def __init__(
        self,
        exp_list=None,
        savedir_base=None,
        workdir=None,
        job_config=None,
        verbose=1,
        account_id=None,
        job_scheduler=None,
        save_logs=True,
    ):
        """[summary]

        Parameters
        ----------
        exp_list : [type]
            [description]
        savedir_base : [type]
            [description]
        workdir : [type], optional
            [description], by default None
        job_config : [type], optional
            [description], by default None
        verbose : int, optional
            [description], by default 1
        """
        job_config = copy.deepcopy(job_config)
        if account_id is None and job_config is not None and "account_id" in job_config:
            account_id = job_config["account_id"]
            del job_config["account_id"]

        self.exp_list = exp_list
        self.job_config = job_config
        self.workdir = workdir
        self.verbose = verbose
        self.savedir_base = savedir_base
        self.account_id = account_id
        self.save_logs = save_logs

        # define funcs
        if job_scheduler == "toolkit":
            from . import toolkit_manager as ho

            self.ho = ho
            self.api = self.ho.get_api(token=None)

        elif job_scheduler == "slurm":
            from . import slurm_manager as ho

            self.ho = ho
            self.api = None

            try:
                ho.sanity_check()
            except Exception as e:
                raise ValueError("slurm is not supported in this machine.")

        else:
            raise ValueError(f'Job Scheduler {job_scheduler} is not defined. Should be one of ["toolkit", "slurm"]')

    def get_command_history(self, topk=10):
        job_list = self.get_jobs()

        count = 0
        for j in job_list:
            if hasattr(j, "command"):
                print(count, ":", j.command[2])
            if count > topk:
                break
            count += 1

    # Base functions
    # --------------
    def get_jobs(self):
        return self.ho.get_jobs(self.api, account_id=self.account_id)

    def get_jobs_dict(self, job_id_list):
        return self.ho.get_jobs_dict(self.api, job_id_list)

    def get_job(self, job_id):
        return self.ho.get_job(self.api, job_id)

    def kill_job(self, job_id):
        return self.ho.kill_job(self.api, job_id)

    def submit_job(self, command, workdir, savedir_logs=None):
        # make sure the paths are absolute
        workdir = os.path.abspath(workdir)
        if savedir_logs is not None:
            savedir_logs = os.path.abspath(savedir_logs)

        return self.ho.submit_job(
            api=self.api,
            account_id=self.account_id,
            command=command,
            job_config=self._adapt_job_config(self.job_config, workdir),
            workdir=workdir,
            savedir_logs=savedir_logs,
        )

    @staticmethod
    def _adapt_job_config(job_config, workdir):
        """Substitute `<special_patterns_like_this>` in job config (e.g., name)

        The input `job_config` is NOT altered in place (i.e., a copy is made if needed).
        """
        # As of right now, substitution is only performed if a str "name" field is provided.
        if isinstance(job_config.get("name", None), str):
            # Get a copy so that we can safely do substitutions.
            job_config = copy.deepcopy(job_config)
            # The <exp_id> substitution is a special case.
            if "<exp_id>" in job_config["name"]:
                # Complication: we don't have `exp_id`, but we have `workdir`.
                savedir = os.path.dirname(workdir)
                assert (
                    os.path.join(savedir, SUFFIX_TO_ADD_TO_SAVEDIR_TO_GET_WORKDIR) == workdir
                ), "Cannot substitute `<exp_id>` because `workdir` has unexpected format."
                fname_exp_dict = os.path.join(savedir, "exp_dict.json")
                exp_dict = hu.load_json(fname_exp_dict)
                exp_id = hu.hash_dict(exp_dict)
                # We may now proceed with the substitution proper for <exp_id>.
                job_config["name"] = job_config["name"].replace("<exp_id>", exp_id)
            # Other substitutions (simpler)
            # Adding time in a name help prevent duplicated names (which are forbidden on toolkit)
            job_config["name"] = job_config["name"].replace("<time>", str(int(time.time())))
            job_config["name"] = job_config["name"].replace("<time_ns>", str(time.time_ns()))
        return job_config

    # Main functions
    # --------------
    def launch_menu(
        self, command=None, exp_list=None, get_logs=False, wait_seconds=3, in_parallel=True, job_option=None
    ):
        exp_list = exp_list or self.exp_list
        summary_list = self.get_summary_list(get_logs=False, exp_list=exp_list)
        summary_dict = hu.group_list(summary_list, key="job_state", return_count=True)

        print("\nTotal Experiments:", len(exp_list))
        print("Experiment Status:", summary_dict)
        prompt = (
            "\nMenu:\n"
            "  0)'ipdb' run ipdb for an interactive session; or\n"
            "  1)'reset' to reset the experiments; or\n"
            "  2)'run' to run the remaining experiments and retry the failed ones; or\n"
            "  3)'status' to view the job status; or\n"
            "  4)'kill' to kill the jobs.\n"
            "Type option: "
        )
        if job_option is None:
            option = input(prompt)
        else:
            option = job_option

        option_list = ["reset", "run", "status", "logs", "kill", "ipdb"]
        if option not in option_list:
            raise ValueError("Prompt input has to be one of these choices %s" % option_list)

        if option == "ipdb":
            import ipdb

            ipdb.set_trace()
            print("Example:\nsummary_dict = self.get_summary(get_logs=True, exp_list=exp_list)")

        elif option == "status":
            # view experiments
            self.print_job_status(exp_list=exp_list)
            return

        elif option == "reset":
            self.verbose = False
            for e_list in chunk_list(exp_list, n=100):
                self.launch_exp_list(command=command, exp_list=e_list, reset=1, in_parallel=in_parallel)

        elif option == "run":
            self.verbose = False
            tmp_list = [
                s_dict["exp_dict"]
                for s_dict in summary_list
                if s_dict["job_state"] not in ALIVE_STATES + COMPLETED_STATES
            ]

            print("Selected %d/%d exps" % (len(tmp_list), len(exp_list)))
            exp_list = tmp_list
            if len(exp_list) == 0:
                print("All experiments have ran.")
                return

            for e_list in chunk_list(exp_list, n=100):
                self.launch_exp_list(command=command, exp_list=e_list, reset=0, in_parallel=in_parallel)

        elif option == "kill":
            self.verbose = False
            tmp_list = [s_dict["exp_dict"] for s_dict in summary_list if s_dict["job_state"] in ALIVE_STATES]
            self.kill_jobs(exp_list=tmp_list)

        # view experiments
        # print("Checking job status...")
        time.sleep(wait_seconds)
        self.print_job_status(exp_list=exp_list)

    def print_job_status(self, exp_list):
        summary_list = self.get_summary_list(get_logs=False, exp_list=exp_list)
        summary_dict = hu.group_list(summary_list, key="job_state", return_count=False)

        for k in summary_dict.keys():
            n_jobs = len(summary_dict[k])
            if n_jobs:
                # print('\nExperiments %s: %d' % (k, n_jobs))
                pass
                # print(pd.DataFrame(summary_dict[k]).head())

        summary_dict = hu.group_list(summary_list, key="job_state", return_count=True)
        print(summary_dict)

    def launch_exp_list(self, command, exp_list=None, savedir_base=None, reset=0, in_parallel=True):
        exp_list = exp_list or self.exp_list
        assert "<exp_id>" in command

        submit_dict = {}

        if in_parallel:
            pr = hu.Parallel()

            for exp_dict in exp_list:
                exp_id = hu.hash_dict(exp_dict)

                savedir_base = savedir_base or self.savedir_base
                savedir = os.path.join(savedir_base, hu.hash_dict(exp_dict))

                com = command.replace("<exp_id>", exp_id)
                pr.add(self.launch_or_ignore_exp_dict, exp_dict, com, reset, savedir, submit_dict)

            pr.run()
            pr.close()

        else:
            for exp_dict in exp_list:
                exp_id = hu.hash_dict(exp_dict)

                savedir_base = savedir_base or self.savedir_base
                savedir = os.path.join(savedir_base, hu.hash_dict(exp_dict))

                com = command.replace("<exp_id>", exp_id)
                self.launch_or_ignore_exp_dict(exp_dict, com, reset, savedir, submit_dict)

        if len(submit_dict) == 0:
            raise ValueError("The threads have an error, most likely a permission error (see above)")

        for i, (k, v) in enumerate(submit_dict.items()):
            print("***")
            print("Exp %d/%d - %s" % (i + 1, len(submit_dict), v["message"]))
            print("exp_id: %s" % hu.hash_dict(v["exp_dict"]))
            print("job_id: %s" % k)
            savedir = os.path.join(savedir_base, hu.hash_dict(v["exp_dict"]))
            print(f"savedir: {savedir}")
            pprint.pprint(v["exp_dict"])
            print()

        print("%d experiments submitted." % len(exp_list))
        if len(submit_dict) > 0:
            assert len(submit_dict) == len(exp_list), "considered exps does not match expected exps"
        return submit_dict

    def kill_jobs(self, exp_list=None):
        exp_list = exp_list or self.exp_list
        hu.check_duplicates(exp_list)

        pr = hu.Parallel()
        submit_dict = {}

        for exp_dict in exp_list:
            exp_id = hu.hash_dict(exp_dict)
            savedir = os.path.join(self.savedir_base, exp_id)
            fname = get_job_fname(savedir)

            if os.path.exists(fname):
                job_id = hu.load_json(fname)["job_id"]
                pr.add(self.kill_job, job_id)
                submit_dict[exp_id] = "KILLED"
            else:
                submit_dict[exp_id] = "NoN-Existent"

        pr.run()
        pr.close()
        pprint.pprint(submit_dict)
        print("%d/%d experiments killed." % (len([s for s in submit_dict.values() if "KILLED" in s]), len(submit_dict)))
        return submit_dict

    def launch_or_ignore_exp_dict(self, exp_dict, command, reset, savedir, submit_dict=None):
        """launch or ignore job.

        It checks if the experiment exist and manages the special casses, e.g.,
        new experiment, reset, failed, job is already running, completed
        """
        # Define paths
        fname = get_job_fname(savedir)

        if not os.path.exists(fname):
            # Check if the job already exists
            job_dict = self.launch_exp_dict(exp_dict, savedir, command, job=None)
            job_id = job_dict["job_id"]
            message = "SUBMITTED: Launching"

        elif reset:
            # Check if the job already exists
            job_id = hu.load_json(fname).get("job_id")
            self.kill_job(job_id)
            hc.delete_and_backup_experiment(savedir)

            job_dict = self.launch_exp_dict(exp_dict, savedir, command, job=None)
            job_id = job_dict["job_id"]
            message = "SUBMITTED: Resetting"

        else:
            job_id = hu.load_json(fname).get("job_id")
            job = self.get_job(job_id)

            if job["state"] in ALIVE_STATES + COMPLETED_STATES:
                # If the job is alive, do nothing
                message = "IGNORED: Job %s" % job["state"]

            elif job["state"] in FAILED_STATES:
                message = "SUBMITTED: Retrying %s Job" % job["state"]
                job_dict = self.launch_exp_dict(exp_dict, savedir, command, job=job)
                job_id = job_dict["job_id"]
            # This shouldn't happen
            else:
                raise ValueError("wtf")

        if submit_dict is None:
            submit_dict = {}

        submit_dict[job_id] = {"exp_dict": exp_dict, "message": message}

        return submit_dict

    def launch_exp_dict(self, exp_dict, savedir, command, job=None):
        """Submit a job job and save job dict and exp_dict."""
        # Check for duplicates
        # if job is not None:
        # assert self._assert_no_duplicates(job)

        fname_exp_dict = os.path.join(savedir, "exp_dict.json")
        hu.save_json(fname_exp_dict, exp_dict)
        exp_id = hu.hash_dict(exp_dict)

        assert hu.hash_dict(hu.load_json(fname_exp_dict)) == exp_id

        # Define paths
        workdir_job = os.path.join(savedir, SUFFIX_TO_ADD_TO_SAVEDIR_TO_GET_WORKDIR)

        # Copy the experiment code into the experiment folder
        print(f"Copying code for experiment {exp_id}")
        hu.copy_code(self.workdir + "/", workdir_job, verbose=0)

        # Run  command
        if self.save_logs:
            savedir_logs = savedir
        else:
            savedir_logs = None
        job_id = self.submit_job(command, workdir_job, savedir_logs=savedir_logs)
        print(f"Job submitted for experiment {exp_id} with job id {job_id}")

        # Verbose
        if self.verbose:
            print("Job_id: %s command: %s" % (job_id, command))

        job_dict = {"job_id": job_id, "command": command}

        hu.save_json(get_job_fname(savedir), job_dict)

        return job_dict

    def get_summary_list(
        self,
        failed_only=False,
        columns=None,
        max_lines=10,
        wrap_size=8,
        add_prefix=False,
        get_logs=True,
        exp_list=None,
        savedir_base=None,
    ):
        savedir_base = savedir_base or self.savedir_base
        exp_list = exp_list or self.exp_list

        # get job key
        job_id_list = []
        for exp_dict in exp_list:
            exp_id = hu.hash_dict(exp_dict)
            savedir = os.path.join(savedir_base, exp_id)
            fname = get_job_fname(savedir)

            if os.path.exists(fname):
                job_id_list += [hu.load_json(fname)["job_id"]]

        jobs_dict = self.get_jobs_dict(job_id_list)

        # get summaries
        summary_list = []

        for exp_dict in exp_list:
            result_dict = {}

            exp_id = hu.hash_dict(exp_dict)
            savedir = os.path.join(savedir_base, exp_id)
            job_fname = get_job_fname(savedir)

            # General info
            result_dict = {}
            result_dict["exp_dict"] = exp_dict
            result_dict["exp_id"] = exp_id
            result_dict["job_id"] = None
            result_dict["job_state"] = "NEVER LAUNCHED"

            if os.path.exists(job_fname):
                job_dict = hu.load_json(job_fname)
                job_id = job_dict["job_id"]
                if job_id not in jobs_dict:
                    continue

                fname_exp_dict = os.path.join(savedir, "exp_dict.json")
                job = jobs_dict[job_id]

                # if hasattr(job, 'command'):
                #     command = job_dict['command']
                # else:
                #     command = None

                # Job info
                result_dict["started_at"] = hu.time_to_montreal(fname_exp_dict)
                result_dict["job_id"] = job_id
                result_dict["job_state"] = job["state"]
                result_dict["restarts"] = len(job["runs"])
                result_dict["command"] = job_dict["command"]

                if get_logs:
                    # Logs info
                    if job["state"] == "FAILED":
                        logs_fname = os.path.join(savedir, "err.txt")
                    else:
                        logs_fname = os.path.join(savedir, "logs.txt")

                    if os.path.exists(logs_fname):
                        result_dict["logs"] = hu.read_text(logs_fname)[-max_lines:]

            summary_list += [result_dict]

        return summary_list

    def get_summary(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        # get job ids

        # fill summary
        summary_dict = {"table": [], "status": [], "logs_failed": [], "logs": []}

        # get info
        df = pd.DataFrame(summary_dict["table"])

        # if columns:
        #     df = df[[c for c in columns if (c in df.columns and c not in ['err'])]]

        if "job_state" in df:
            stats = np.vstack(np.unique(df["job_state"].fillna("NaN"), return_counts=True)).T
            status = [{a: b} for (a, b) in stats]
        else:
            df["job_state"] = None

        df = hu.sort_df_columns(df)
        summary_dict["status"] = status
        summary_dict["table"] = df

        for state in ALIVE_STATES + COMPLETED_STATES + FAILED_STATES:
            summary_dict[state] = df[df["job_state"] == state]

        return summary_dict

    def _assert_no_duplicates(self, job_new=None, max_jobs=500):
        # Get the job list
        jobList = self.get_jobs()

        # Check if duplicates already exist in job
        command_dict = {}
        for job in jobList:

            if hasattr(job, "command"):
                if job.command is None:
                    continue
                job_python_command = job.command[2]
            else:
                job_python_command = None

            if job_python_command is None:
                continue
            elif job_python_command not in command_dict:
                command_dict[job_python_command] = job
            else:
                print("Job state", job["state"], "Job command", job_python_command)
                raise ValueError("Job %s is duplicated" % job_python_command)

        # Check if the new job causes duplicate
        if job_new is not None:
            if job_new.command[2] in command_dict:
                job_old_id = command_dict[job_new.command[2]].id
                raise ValueError("Job exists as %s" % job_old_id)

        return True


def get_job_fname(savedir):
    if os.path.exists(os.path.join(savedir, "borgy_dict.json")):
        # for backward compatibility
        fname = os.path.join(savedir, "borgy_dict.json")
    else:
        fname = os.path.join(savedir, "job_dict.json")

    return fname


def chunk_list(my_list, n=100):
    return [my_list[x : x + n] for x in range(0, len(my_list), n)]


def launch_job(command, savedir_base, job_scheduler, job_config, reset=False, verbose=True):
    exp_dict = {"command": command}

    jm = JobManager(
        exp_list=[exp_dict],
        savedir_base=savedir_base,
        workdir=os.getcwd(),
        job_config=job_config,
        job_scheduler=job_scheduler,
        save_logs=True,
    )
    exp_id = hu.hash_dict(exp_dict)
    savedir = f"{savedir_base}/{exp_id}"
    fname = get_job_fname(savedir)

    if not os.path.exists(fname):
        message = f"No Job"
    else:
        job_id = hu.load_json(fname).get("job_id")
        job = jm.get_job(job_id)
        message = f"Job {job['id']} - {job['state']}"

    print(message)
    print("command:", command)
    print(f"saved: {savedir}")
    prompt = "\nMenu:\n" "  0)'log'; or\n" "  1)'run'; or\n" "  2)'reset'; or\n" "  3)'exit'.\n" "Type option: "

    while 1:
        option = input(prompt)
        if option in ["reset", "2"]:
            reset = 1
            break
        elif option in ["run", "1"]:
            reset = 0
            break
        elif option in ["log", "0"]:
            # Logs info
            if job["state"] == "FAILED":
                logs_fname = os.path.join(savedir, "err.txt")
            else:
                logs_fname = os.path.join(savedir, "logs.txt")

            if os.path.exists(logs_fname):
                logs = hu.read_text(logs_fname)[-10:]
                print(logs)
                print()
                print(message)

        elif option in ["exit", "3"]:
            print("exiting...")
            break
        else:
            raise ValueError(f"{option} does not exist.")

    if option in ["run", "reset"]:
        job = jm.launch_or_ignore_exp_dict(exp_dict=exp_dict, reset=reset, savedir=savedir, command=command)

        if verbose:
            job_id = list(job.keys())[0]
            pprint.pprint({"job_id": job_id, "savedir": savedir, "command": command, "message": job[job_id]["message"]})

        return job



import os
import sys
import torch
import argparse
import pandas as pd
import pprint
import haven

from . import haven_utils as hu
from . import haven_jupyter as hj
import numpy as np


def get_args():
    parser = argparse.ArgumentParser(formatter_class=make_wide(argparse.ArgumentDefaultsHelpFormatter))

    parser.add_argument("-e", "--exp_group_list", nargs="+", help="Define which exp groups to run.")
    parser.add_argument(
        "-sb", "--savedir_base", default=None, help="Define the base directory where the experiments will be saved."
    )
    parser.add_argument("-r", "--reset", default=0, type=int, help="Reset or resume the experiment.")
    parser.add_argument("--exp_id", default=None, help="Run a specific experiment based on its id.")
    parser.add_argument(
        "-j", "--job_scheduler", default=None, type=str, help="Run the experiments as jobs in the cluster."
    )

    args, others = parser.parse_known_args()

    return args


def make_wide(formatter, w=120, h=36):
    """Return a wider HelpFormatter, if possible."""
    try:
        # https://stackoverflow.com/a/5464440
        # beware: "Only the name of this class is considered a public API."
        kwargs = {"width": w, "max_help_position": h}
        formatter(None, **kwargs)
        return lambda prog: formatter(prog, **kwargs)
    except TypeError:
        import warnings

        warnings.warn("argparse help formatter failed, falling back.")
        return formatter


def run_wizard(
    func,
    exp_list=None,
    exp_groups=None,
    job_config=None,
    savedir_base=None,
    reset=None,
    args=None,
    use_threads=False,
    exp_id=None,
    python_binary_path="python",
    python_file_path=None,
    workdir=None,
    job_scheduler=None,
    save_logs=True,
    filter_duplicates=False,
    results_fname=None,
    job_option=None,
):
    if args is None:
        args = get_args()
        custom_args = {}
    else:
        custom_args = vars(args).copy()
        for k, v in vars(get_args()).items():
            if k in custom_args:
                continue
            setattr(args, k, v)

    # Asserts
    # =======
    savedir_base = savedir_base or args.savedir_base
    reset = reset or args.reset
    exp_id = exp_id or args.exp_id
    assert savedir_base is not None

    # make sure savedir_base is absolute
    savedir_base = os.path.abspath(savedir_base)

    # Collect experiments
    # ===================
    if exp_id is not None:
        # select one experiment
        savedir = os.path.join(savedir_base, exp_id)
        exp_dict = hu.load_json(os.path.join(savedir, "exp_dict.json"))

        exp_list = [exp_dict]

    elif exp_list is None:
        # select exp group
        exp_list = []
        for exp_group_name in args.exp_group_list:
            exp_list += exp_groups[exp_group_name]

    if filter_duplicates:
        n_total = len(exp_list)
        exp_list = hu.filter_duplicates(exp_list)
        print(f"Filtered {len(exp_list)}/{n_total}")

    hu.check_duplicates(exp_list)
    print("\nRunning %d experiments" % len(exp_list))

    # save results folder
    if exp_id is None and results_fname is not None:
        if len(results_fname):
            if ".ipynb" not in results_fname:
                raise ValueError(".ipynb should be the file extension")
            hj.create_jupyter_file(fname=results_fname, savedir_base=savedir_base)

    # Run experiments
    # ===============
    if job_scheduler is None:
        job_scheduler = args.job_scheduler

    if job_scheduler in [None, "None", "0"]:
        job_scheduler = None

    elif job_scheduler in ["toolkit", "slurm", "gcp"]:
        job_scheduler = job_scheduler

    elif job_scheduler in ["1"]:
        job_scheduler = "toolkit"

    else:
        raise ValueError(f"{job_scheduler} does not exist")

    if job_scheduler is None:
        for exp_dict in exp_list:
            savedir = create_experiment(exp_dict, savedir_base, reset=reset, verbose=True)
            # do trainval
            func(exp_dict=exp_dict, savedir=savedir, args=args)

    else:
        # launch jobs
        print(f"Using Job Scheduler: {job_scheduler}")

        from haven import haven_jobs as hjb

        assert job_config is not None
        assert "account_id" in job_config

        if workdir is None:
            workdir = os.getcwd()

        jm = hjb.JobManager(
            exp_list=exp_list,
            savedir_base=savedir_base,
            workdir=workdir,
            job_config=job_config,
            job_scheduler=job_scheduler,
            save_logs=save_logs,
        )

        if python_file_path is None:
            python_file_path = os.path.split(sys.argv[0])[-1]

        command = f"{python_binary_path} {python_file_path} --exp_id <exp_id> --savedir_base {savedir_base}"

        for k, v in custom_args.items():
            if k not in [
                "savedir_base",
                "sb",
                "exp_id",
                "e",
                "exp_group_list",
                "j",
                "job_scheduler",
                "r",
                "reset",
            ]:
                command += f" --{k} {v}"

        print(command)
        jm.launch_menu(command=command, in_parallel=use_threads, job_option=job_option)


def create_experiment(exp_dict, savedir_base, reset, copy_code=False, return_exp_id=False, verbose=True):
    import pprint
    from . import haven_chk as hc

    exp_id = hu.hash_dict(exp_dict)
    savedir = os.path.join(savedir_base, exp_id)

    if reset:
        hc.delete_and_backup_experiment(savedir)

    # create experiment structure
    os.makedirs(savedir, exist_ok=True)

    # -- save exp_dict only when it is needed
    exp_dict_json_fname = os.path.join(savedir, "exp_dict.json")
    if not os.path.exists(exp_dict_json_fname):
        hu.save_json(exp_dict_json_fname, exp_dict)
    else:
        # make sure it is not corrupt and same exp_id
        try:
            exp_dict_tmp = hu.load_json(exp_dict_json_fname)
            assert hu.hash_dict(exp_dict_tmp) == hu.hash_dict(exp_dict)
        except:
            hu.save_json(exp_dict_json_fname, exp_dict)

    # -- images
    os.makedirs(os.path.join(savedir, "images"), exist_ok=True)

    if copy_code:
        src = os.getcwd() + "/"
        dst = os.path.join(savedir, "code")
        hu.copy_code(src, dst)

    if verbose:
        print("\n******")
        print(f"Haven: {haven.__version__}")
        print("Exp id: %s" % exp_id)
        print("\nHyperparameters:\n" + "-" * 16)
        # print(pd.DataFrame([exp_dict]).to_string(index=False))
        pprint.pprint(exp_dict)

        print("\nSave directory: %s" % savedir)
        print("=" * 100)

    if return_exp_id:
        return savedir, exp_id

    return savedir


class CheckpointManager:
    def __init__(self, savedir, verbose=True):
        self.savedir = savedir
        self.verbose = verbose
        self.chk_dict = get_checkpoint(savedir)

    def log_metrics(self, score_dict):
        self.chk_dict["score_list"] += [score_dict]
        hu.save_pkl(os.path.join(self.savedir, "score_list.pkl"), self.chk_dict["score_list"])
        if self.verbose:
            report(self.savedir, self.chk_dict["score_list"])

    def load_model(self):
        fname = os.path.join(self.savedir, "model.pth")
        if os.path.exists(fname):
            return torch.load(fname)
        else:
            return None

    def save_model(state_dict):
        hu.torch_save(os.path.join(self.savedir, "model.pth"), state_dict)

    def get_epoch(self):
        if len(self.chk_dict["score_list"]) == 0:
            return 0
        return self.chk_dict["score_list"][-1]["epoch"] + 1

    def save_pkl(self, fname, data):
        hu.save_pkl(os.path.join(self.savedir, fname), data)

    def load_pkl(self, fname, data):
        hu.load_pkl(os.path.join(self.savedir, fname))

    def save_torch(self, fname, data):
        hu.torch_save(os.path.join(self.savedir, fname), data)

    def load_torch(self, fname, data):
        torch.load(os.path.join(self.savedir, fname))

    def save_images(self, images):
        if not isinstance(images, list):
            images = [images]

        for im in images:
            hu.save_image(os.path.join(self.savedir, "images", f"{i}.jpg"), im)


def save_checkpoint(
    savedir, score_list, model_state_dict=None, images=None, images_fname=None, fname_suffix="", verbose=True
):
    # Report
    if verbose:
        report(savedir, score_list)

    # save score_list
    score_list_fname = os.path.join(savedir, "score_list%s.pkl" % fname_suffix)
    hu.save_pkl(score_list_fname, score_list)
    # if verbose:
    # print('> Saved "score_list" as %s' %
    #       os.path.split(score_list_fname)[-1])

    # save model
    if model_state_dict is not None:
        model_state_dict_fname = os.path.join(savedir, "model%s.pth" % fname_suffix)
        hu.torch_save(model_state_dict_fname, model_state_dict)
        # if verbose:
        # print('> Saved "model_state_dict" as %s' %
        #       os.path.split(model_state_dict_fname)[-1])

    # save images
    images_dir = os.path.join(savedir, "images%s" % fname_suffix)
    if images is not None:
        for i, img in enumerate(images):

            if images_fname is not None:
                fname = "%s" % images_fname[i]
            else:
                fname = "%d.png" % i
            hu.save_image(os.path.join(images_dir, fname), img)
        # if verbose:
        #     print('> Saved "images" in %s' % os.path.split(images_dir)[-1])

    print("=" * 100)


def report(savedir, score_list):
    exp_dict = hu.load_json(os.path.join(savedir, "exp_dict.json"))
    print("\nExp id: %s" % hu.hash_dict(exp_dict))

    print("\nHyperparameters:\n" + "-" * 16)
    # print(pd.DataFrame([exp_dict]).to_string(index=False))
    pprint.pprint(exp_dict)
    print("\nMetrics:\n" + "-" * 8)
    score_df = pd.DataFrame(score_list)
    print(score_df.tail().to_string(index=False), "\n")

    print("Save directory: %s" % savedir)


def get_checkpoint(savedir, return_model_state_dict=False):
    chk_dict = {}

    # score list
    score_list_fname = os.path.join(savedir, "score_list.pkl")
    if os.path.exists(score_list_fname):
        score_list = hu.load_pkl(score_list_fname)
    else:
        score_list = []

    chk_dict["score_list"] = score_list
    if len(score_list) == 0:
        chk_dict["epoch"] = 0
    else:
        chk_dict["epoch"] = score_list[-1]["epoch"] + 1

    chk_dict["model_state_dict"] = {}
    model_state_dict_fname = os.path.join(savedir, "model.pth")
    if return_model_state_dict:
        if os.path.exists(model_state_dict_fname):
            chk_dict["model_state_dict"] = hu.torch_load(model_state_dict_fname)
    return chk_dict



import pandas as pd
import pprint
from haven.haven_jobs import slurm_manager as sm
from haven import haven_utils as hu
from haven import haven_results as hr
from haven.haven_results import plots_line as pl
import sys
import os
import pprint

path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0, path)


def get_latex_table(
    score_df, columns=None, rows=None, filter_dict=dict(), map_row_dict_dict=dict(), map_col_dict=dict(), **kwargs
):

    # break it
    dicts = score_df.T.to_dict()

    dicts_new = {}
    for i in dicts:
        exp_score_dict = dicts[i]
        row_label = "_".join(
            [map_row_dict_dict.get(exp_score_dict[l], exp_score_dict[l]) for l in exp_score_dict if l in rows]
        )
        col_scores = {k: v for k, v in exp_score_dict.items() if k in columns}
        dicts_new[row_label] = col_scores

    df_new = pd.DataFrame(dicts_new).T
    return df_new.to_latex(**kwargs)


if __name__ == "__main__":
    savedir_base = ".tmp"
    exp_dict = {"model": {"name": "mlp", "n_layers": 30}, "dataset": "mnist", "batch_size": 1}

    score_list = [{"epoch": 4, "acc": 0.5}, {"epoch": 6, "acc": 0.9}]

    hu.save_pkl(os.path.join(savedir_base, hu.hash_dict(exp_dict), "score_list.pkl"), score_list)

    hu.save_json(os.path.join(savedir_base, hu.hash_dict(exp_dict), "exp_dict.json"), exp_dict)

    exp_dict = {"model": {"name": "mlp2", "n_layers": 35}, "dataset": "mnist", "batch_size": 1}
    score_list = [{"epoch": 2, "acc": 0.1}, {"epoch": 6, "acc": 0.3}]
    hu.save_pkl(os.path.join(savedir_base, hu.hash_dict(exp_dict), "score_list.pkl"), score_list)
    hu.save_json(os.path.join(savedir_base, hu.hash_dict(exp_dict), "exp_dict.json"), exp_dict)
    # check if score_list can be loaded and viewed in pandas
    exp_list = hu.get_exp_list(savedir_base=savedir_base)
    score_df = hr.get_score_df(exp_list, savedir_base=savedir_base)

    print(
        get_latex_table(
            score_df,
            columns=["acc", "epoch"],
            rows=["model.name"],
            filter_dict=dict(),
            caption="Test",
            float_format="%.2f",
            label="ref:table1",
            map_row_dict_dict=dict(),
            map_col_dict=dict(),
        )
    )
    print()



import ast


def get_dict_from_str(string):
    if string is None:
        return string

    if string == "None":
        return None

    if string == "":
        return None

    return ast.literal_eval(string)


def get_list_from_str(string):
    if string is None:
        return string

    if string == "None":
        return None

    string = string.replace(" ", "").replace("]", "").replace("[", "").replace('"', "").replace("'", "")

    if string == "":
        return None

    return string.split(",")



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu
from .. import haven_share as hd
from . import widgets as wdg

import os
import pprint
import json
import copy
import pprint
import pandas as pd

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except Exception:
    print("widgets not available...")


def latex_tab(db, output):

    b_table = widgets.Button(description="Display Latex Table")

    w_rows = wdg.SelectMultiple(header="Legend", options=db.rm.exp_params, db_vars=db.vars, var="latex_rows")
    w_cols = wdg.SelectMultiple(header="Metrics", options=db.rm.score_keys, db_vars=db.vars, var="latex_columns")

    button = widgets.VBox(
        [
            widgets.HBox([w_rows.get_widget(), w_cols.get_widget()]),
            widgets.HBox([b_table]),
        ]
    )
    output_plot = widgets.Output()

    with output:
        display(button)
        display(output_plot)

    def on_clicked(b):
        output_plot.clear_output()
        with output_plot:
            db.update_rm()

            score_table = db.rm.get_latex_table(legend=w_cols.update(), metrics=w_rows.update(), caption="Results")
            print(score_table)

    b_table.on_click(on_clicked)



import os
import json
import copy
import sys
import hashlib
import itertools
from itertools import groupby
import tqdm
import pprint
import numpy as np
from .. import haven_utils as hu


def cartesian_exp_group(exp_config, remove_none=False):
    """Cartesian experiment config.

    It converts the exp_config into a list of experiment configuration by doing
    the cartesian product of the different configuration. It is equivalent to
    do a grid search.

    Parameters
    ----------
    exp_config : str
        Dictionary with the experiment Configuration

    Returns
    -------
    exp_list: str
        A list of experiments, each defines a single set of hyper-parameters
    """
    exp_config_copy = copy.deepcopy(exp_config)

    # Make sure each value is a list
    for k, v in exp_config_copy.items():
        if not isinstance(exp_config_copy[k], list):
            exp_config_copy[k] = [v]

    # Create the cartesian product
    exp_list_raw = (
        dict(zip(exp_config_copy.keys(), values)) for values in itertools.product(*exp_config_copy.values())
    )

    # Convert into a list
    exp_list = []
    for exp_dict in exp_list_raw:
        # remove hparams with None
        if remove_none:
            to_remove = []
            for k, v in exp_dict.items():
                if v is None:
                    to_remove += [k]
            for k in to_remove:
                del exp_dict[k]
        exp_list += [exp_dict]

    return exp_list


def _filter_fn(val):
    if val == "true":
        return True
    elif val == "false":
        return False
    elif val == "none":
        return None
    else:
        return val


def _traverse(dict_, keys, val):
    if len(keys) == 0:
        return
    else:
        # recurse
        if keys[0] not in dict_:
            if len(keys[1:]) == 0:
                dict_[keys[0]] = _filter_fn(val)
            else:
                dict_[keys[0]] = {}
        _traverse(dict_[keys[0]], keys[1:], val)


def get_exp_list_from_json(fname=None, json_dict=None):
    if fname is not None:
        json_dict = json.loads(open(fname).read())

    exps = cartesian_exp_group(json_dict)
    return [unflatten(dd) for dd in exps]


def unflatten(dict_):
    new_dict = {}
    for key, val in dict_.items():
        key_split = key.split(".")
        _traverse(new_dict, key_split, val)
    return new_dict


def hash_dict(exp_dict):
    """Create a hash for an experiment.

    Parameters
    ----------
    exp_dict : dict
        An experiment, which is a single set of hyper-parameters

    Returns
    -------
    hash_id: str
        A unique id defining the experiment
    """
    dict2hash = ""
    if not isinstance(exp_dict, dict):
        raise ValueError("exp_dict is not a dict")

    for k in sorted(exp_dict.keys()):
        if "." in k:
            raise ValueError(". has special purpose")
        elif isinstance(exp_dict[k], dict):
            v = hash_dict(exp_dict[k])
        elif isinstance(exp_dict[k], tuple):
            raise ValueError(f"{exp_dict[k]} tuples can't be hashed yet, consider converting tuples to lists")
        elif isinstance(exp_dict[k], list) and len(exp_dict[k]) and isinstance(exp_dict[k][0], dict):
            v_str = ""
            for e in exp_dict[k]:
                if isinstance(e, dict):
                    v_str += hash_dict(e)
                else:
                    raise ValueError("all have to be dicts")
            v = v_str
        else:
            v = exp_dict[k]

        dict2hash += str(k) + "/" + str(v)
    hash_id = hashlib.md5(dict2hash.encode()).hexdigest()

    return hash_id


def hash_str(str):
    """Hashes a string using md5.

    Parameters
    ----------
    str
        a string

    Returns
    -------
    str
        md5 hash for the input string
    """
    return hashlib.md5(str.encode()).hexdigest()


def group_exp_list(exp_list, groupby_list):
    """Split the experiment list into smaller lists where each
       is grouped by a set of hyper-parameters

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    groupby_list : list
        The set of hyper-parameters to group the experiments

    Returns
    -------
    lists_of_exp_list : list
        Experiments grouped by a set of hyper-parameters

    Example
    -------
    >>>
    >>>
    >>>
    """
    if groupby_list is None:
        return [exp_list]
    if not isinstance(groupby_list, list):
        groupby_list = [groupby_list]
    # groupby_list = hu.as_double_list(groupby_list)
    exp_list = copy.deepcopy(exp_list)
    exp_list_flat = [hu.flatten_column(exp_dict, flatten_list=True) for exp_dict in exp_list]
    for e1, e2 in zip(exp_list_flat, exp_list):
        e1["exp_dict"] = e2

    def split_func(x):
        x_list = []
        for k_list in groupby_list:
            if not isinstance(k_list, list):
                k_list = [k_list]
            val = get_str(x, k_list)
            x_list += [val]

        return x_list

    exp_list_flat.sort(key=split_func)

    list_of_exp_list = []
    group_dict = groupby(exp_list_flat, key=split_func)

    # exp_group_dict = {}
    for k, v in group_dict:
        v_list = [vi["exp_dict"] for vi in list(v)]
        list_of_exp_list += [v_list]
    #     # print(k)
    #     exp_group_dict['_'.join(list(map(str, k)))] = v_list

    return list_of_exp_list


def get_exp_diff(exp_list):
    flat_list = [hu.flatten_column(e, flatten_list=True) for e in exp_list]

    hparam_dict = {}
    for f in flat_list:
        for k, v in f.items():
            if k not in hparam_dict:
                hparam_dict[k] = set()
            hparam_dict[k].add(v)
    count_dict = []
    for k, v in hparam_dict.items():
        count_dict += [{"name": k, "uniques": len(v)}]

    count_dict = sorted(count_dict, key=lambda i: -i["uniques"])
    return count_dict


def group_list(python_list, key, return_count=False):
    group_dict = {}
    for p in python_list:
        p_tmp = copy.deepcopy(p)
        del p_tmp[key]
        k = p[key]

        if k not in group_dict:
            group_dict[k] = []

        group_dict[k] += [p_tmp]

    if return_count:
        count_dict = {}
        for k in group_dict:
            count_dict[k] = len(group_dict[k])
        return count_dict
    return group_dict


def get_exp_list_from_config(exp_groups, exp_config_fname):
    exp_list = []
    for e in exp_groups:
        exp_list += hu.load_py(exp_config_fname).EXP_GROUPS[e]

    return exp_list


def get_str(h_dict, k_list):
    k = k_list[0]

    if len(k_list) == 1:
        return str(h_dict.get(k))

    return get_str(h_dict.get(k), k_list[1:])


def get_best_exp_dict(
    exp_list,
    savedir_base,
    metric,
    metric_agg="min",
    filterby_list=None,
    avg_across=None,
    return_scores=False,
    verbose=True,
    score_list_name="score_list.pkl",
    return_score_dict=False,
):
    """Obtain best the experiment for a specific metric.

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    savedir_base : [type]
        A directory where experiments are saved
    metric : [type]
        [description]
    min_or_max : [type]
        [description]
    return_scores : bool, optional
        [description], by default False
    """
    scores_dict = []
    if metric_agg in ["min", "min_last"]:
        best_score = np.inf
    elif metric_agg in ["max", "max_last"]:
        best_score = 0.0

    exp_dict_best = None
    best_score_dict = None
    exp_list = filter_exp_list(exp_list, filterby_list, verbose=verbose)
    for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
        savedir = os.path.join(savedir_base, exp_id)

        score_list_fname = os.path.join(savedir, score_list_name)
        if not os.path.exists(score_list_fname):
            if verbose:
                print("%s: missing %s" % (exp_id, score_list_name))
            continue

        score_list = hu.load_pkl(score_list_fname)
        metric_scores = [score_dict[metric] for score_dict in score_list if metric in score_dict]
        if len(metric_scores) == 0:
            continue
        if metric_agg in ["min", "min_last"]:
            if metric_agg == "min_last":
                score = metric_scores[-1]
            elif metric_agg == "min":
                score = np.min(metric_scores)

            if best_score >= score:
                best_score = score
                exp_dict_best = exp_dict
                best_score_dict = {"score": score, "epochs": len(score_list), "exp_id": exp_id}

        elif metric_agg in ["max", "max_last"]:
            if metric_agg == "max_last":
                score = metric_scores[-1]
            elif metric_agg == "max":
                score = np.max(metric_scores)

            if best_score <= score:
                best_score = score
                exp_dict_best = exp_dict
                best_score_dict = {"score": score, "epochs": len(score_list), "exp_id": exp_id}

    if exp_dict_best is None:
        if verbose:
            print('no experiments with metric "%s"' % metric)
        return {}

    if return_score_dict:
        return exp_dict_best, best_score_dict

    return exp_dict_best


def get_exp_list_from_exp_configs(exp_group_list, workdir, filterby_list=None, verbose=True):
    """[summary]

    Parameters
    ----------
    exp_group_list : [type]
        [description]
    workdir : [type]
        [description]
    filterby_list : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    assert workdir is not None

    from importlib import reload

    assert workdir is not None
    if workdir not in sys.path:
        sys.path.append(workdir)
    import exp_configs as ec

    reload(ec)

    exp_list = []
    for exp_group in exp_group_list:
        exp_list += ec.EXP_GROUPS[exp_group]
    if verbose:
        print("%d experiments" % len(exp_list))

    exp_list = filter_exp_list(exp_list, filterby_list, verbose=verbose)
    return exp_list


def get_exp_ids(savedir_base, filterby_list=None, verbose=True):
    return get_exp_list(savedir_base, filterby_list, verbose, return_ids=True)


def get_exp_list(savedir_base, filterby_list=None, verbose=True, return_ids=False):
    """[summary]

    Parameters
    ----------
    savedir_base : [type], optional
        A directory where experiments are saved, by default None
    filterby_list : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    exp_list = []
    dir_list = os.listdir(savedir_base)

    for exp_id in tqdm.tqdm(dir_list):
        savedir = os.path.join(savedir_base, exp_id)
        fname = os.path.join(savedir, "exp_dict.json")
        if len(exp_id) != 32:
            if verbose:
                print('"%s/" is not an exp directory' % exp_id)
            continue

        if not os.path.exists(fname):
            if verbose:
                print("%s: missing exp_dict.json" % exp_id)
            continue

        exp_dict = hu.load_json(fname)
        expected_id = hu.hash_dict(exp_dict)
        if expected_id != exp_id:
            if verbose:
                # assert(hu.hash_dict(exp_dict) == exp_id)
                print("%s does not match %s" % (expected_id, exp_id))
            continue
        # print(hu.hash_dict(exp_dict))
        exp_list += [exp_dict]

    exp_list = filter_exp_list(exp_list, filterby_list)
    if return_ids:
        return [hu.hash_dict(e) for e in exp_list]
    return exp_list


def zip_exp_list(savedir_base):
    """[summary]

    Parameters
    ----------
    savedir_base : [type]
        [description]
    """
    import zipfile

    with zipfile.ZipFile(savedir_base) as z:
        for filename in z.namelist():
            if not os.path.isdir(filename):
                # read the file
                with z.open(filename) as f:
                    for line in f:
                        print(line)


def filter_list(python_list, filterby_list, verbose=True):
    return filter_exp_list(python_list, filterby_list, verbose=verbose)


def filter_exp_list(
    exp_list, filterby_list, savedir_base=None, verbose=True, score_list_name="score_list.pkl", return_style_list=False
):
    """[summary]

    Parameters
    ----------
    exp_list : [type]
        A list of experiments, each defines a single set of hyper-parameters
    filterby_list : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    if filterby_list is None or filterby_list == "" or len(filterby_list) == 0:
        if return_style_list:
            return exp_list, [{}] * len(exp_list)
        else:
            return exp_list

    style_list = []
    filterby_list_list = hu.as_double_list(filterby_list)
    # filterby_list = filterby_list_list

    for filterby_list in filterby_list_list:
        exp_list_new = []

        # those with meta
        filterby_list_no_best = []
        for filterby_dict in filterby_list:
            meta_dict = {}
            if isinstance(filterby_dict, tuple):
                fd, meta_dict = filterby_dict

            if meta_dict.get("best"):
                assert savedir_base is not None
                el = filter_exp_list(exp_list, filterby_list=fd, verbose=verbose)
                best_dict = meta_dict.get("best")
                exp_dict = get_best_exp_dict(
                    el,
                    savedir_base,
                    metric=best_dict["metric"],
                    metric_agg=best_dict["metric_agg"],
                    filterby_list=None,
                    avg_across=best_dict.get("avg_across"),
                    return_scores=False,
                    verbose=verbose,
                    score_list_name=score_list_name,
                )

                exp_list_new += [exp_dict]
                style_list += [meta_dict.get("style", {})]
            else:
                filterby_list_no_best += [filterby_dict]

        # ignore metas here meta
        for exp_dict in exp_list:
            select_flag = False

            for fd in filterby_list_no_best:
                if isinstance(fd, tuple):
                    filterby_dict, meta_dict = fd
                    style_dict = meta_dict.get("style", {})
                else:
                    filterby_dict = fd
                    style_dict = {}

                filterby_dict = copy.deepcopy(filterby_dict)

                keys = filterby_dict.keys()

                tmp_dict = {}
                for k in keys:
                    if "." in k:
                        v = filterby_dict[k]
                        k_list = k.split(".")
                        nk = len(k_list)

                        dict_tree = dict()
                        t = dict_tree

                        for i in range(nk):
                            ki = k_list[i]
                            if i == (nk - 1):
                                t = t.setdefault(ki, v)
                            else:
                                t = t.setdefault(ki, {})

                        tmp_dict.update(dict_tree)
                    else:
                        tmp_dict[k] = filterby_dict[k]

                filterby_dict = tmp_dict
                assert isinstance(filterby_dict, dict), "filterby_dict: %s is not a dict" % str(filterby_dict)

                if hu.is_subset(filterby_dict, exp_dict):
                    select_flag = True
                    break

            if select_flag:
                exp_list_new += [exp_dict]
                style_list += [style_dict]

        exp_list = exp_list_new

    if verbose:
        print("Filtered: %d/%d experiments gathered..." % (len(exp_list_new), len(exp_list)))
    # hu.check_duplicates(exp_list_new)
    exp_list_new = hu.ignore_duplicates(exp_list_new)

    if return_style_list:
        return exp_list_new, style_list

    return exp_list_new



import numpy as np
import copy
import pycocotools.mask as mask_util
from itertools import product
import torch


class APMonitor:
    def __init__(self):
        self.pred_ann_list = []
        self.gt_ann_list = []
        self.n_batches = 0.0
        self.iou_thr = 0.5
        self.iou_thr_list = [0.5, 0.75]

    def add(self, gt_ann_list, pred_ann_list):
        self.pred_ann_list += pred_ann_list
        self.gt_ann_list += gt_ann_list

        self.n_batches += 1.0

    def get_avg_score(self):
        results = compute_precision(
            self.gt_ann_list, self.pred_ann_list, iouType="segm", iouThr=0.5, iouThrList=np.array([0.25, 0.5, 0.75])
        )
        results = {k: v for k, v in results.items() if "mAP" in k}
        results["val_score"] = results["mAP50.0"]

        return results


def compute_precision(gt_annList, pred_annList, iouType, iouThr, iouThrList):
    gt_annList = copy.deepcopy(gt_annList)
    # pred_annList = copy.deepcopy(pred_annList)
    if len(gt_annList) == 0:
        return {"mAP50.0": 0}
    if len(pred_annList) == 0:
        return {"mAP50.0": 0}

    result_dict = evaluate_annList(
        pred_annList=pred_annList, gt_annList=gt_annList, iouType=iouType, iouThr=iouThr, ap=1, iouThrList=iouThrList
    )

    return result_dict


def evaluate_annList(
    pred_annList, gt_annList, ap=1, iouType="bbox", iouThr=0.5, maxDets=100, aRngLabel="all", iouThrList=None
):
    if iouThrList is None:
        iouThrList = np.linspace(0.5, 0.95, int(np.round((0.95 - 0.5) / 0.05)) + 1, endpoint=True)

    # print("iouThrList", iouThrList)
    recThrList = np.linspace(0.0, 1.00, int(np.round((1.00 - 0.0) / 0.01)) + 1, endpoint=True)

    areaRngList = [[0 ** 2, 1e5 ** 2], [0 ** 2, 32 ** 2], [32 ** 2, 96 ** 2], [96 ** 2, 1e5 ** 2]]
    aRngLabelList = ["all", "small", "medium", "large"]
    maxDetList = [1, 10, 100]

    # Put them in dict
    gt_dict = {}  # gt for evaluation
    dt_dict = {}  # dt for evaluation
    imgList = set()
    catList = set()

    for i, gt in enumerate(copy.deepcopy(gt_annList)):
        key = (gt["image_id"], gt["category_id"])
        if key not in gt_dict:
            gt_dict[key] = []

        gt["id"] = i + 1
        gt_dict[key] += [gt]

        imgList.add(gt["image_id"])
        catList.add(gt["category_id"])

    for i, dt in enumerate(copy.deepcopy(pred_annList)):
        key = (dt["image_id"], dt["category_id"])
        if key not in dt_dict:
            dt_dict[key] = []

        bb = dt["bbox"]
        dt["area"] = bb[2] * bb[3]

        dt["id"] = i + 1
        dt["iscrowd"] = 0

        dt_dict[key] += [dt]

        # imgList.add(dt['image_id'])
        # catList.add(dt['category_id'])

    imgList = list(imgList)
    catList = list(catList)

    # compute ious
    iou_dict = {}
    for imgId, catId in product(imgList, catList):
        key = (imgId, catId)

        if key in gt_dict:
            gt = gt_dict[key]
        else:
            gt = []

        if key in dt_dict:
            dt = dt_dict[key]
        else:
            dt = []

        iou_dict[key] = computeIoU(gt, dt, iouType=iouType, maxDets=maxDets)
    # evaluate detections
    evalImgs = []
    for catId, areaRng, imgId in product(catList, areaRngList, imgList):
        evalImgs += [evaluateImg(gt_dict, dt_dict, iou_dict, imgId, catId, areaRng, maxDets, iouThrs=iouThrList)]

    eval_dict = accumulate(evalImgs, imgList, catList, iouThrList, recThrList, areaRngList, maxDetList)

    result_dict = {"iouThrList": iouThrList, "iouType": iouType, "n_classes": len(catList), "categories": catList}

    # Get thresholds for individual APs
    aind = [i for i, aRng in enumerate(aRngLabelList) if aRng == aRngLabel]
    mind = [i for i, mDet in enumerate(maxDetList) if mDet == maxDets]

    for thr in iouThrList:
        ap_dict = compute_AP_dict(ap, aind, mind, eval_dict, thr, iouThrList)
        result_dict.update(ap_dict)

    if ap == 1:
        result_dict["mAP"] = result_dict["mAP%s" % (iouThr * 100)]
    else:
        result_dict["mRC"] = result_dict["mRC%s" % (iouThr * 100)]
    return result_dict


def computeIoU(gt, dt, iouType="bbox", maxDets=100):
    if len(gt) == 0 and len(dt) == 0:
        return []

    inds = np.argsort([-d["score"] for d in dt], kind="mergesort")
    dt = [dt[i] for i in inds]
    if len(dt) > maxDets:
        dt = dt[0:maxDets]

    if iouType == "segm":
        g = [g["segmentation"] for g in gt]
        d = [d["segmentation"] for d in dt]
    elif iouType == "bbox":
        g = [g["bbox"] for g in gt]
        d = [d["bbox"] for d in dt]
    else:
        raise Exception("unknown iouType for iou computation")

    # compute iou between each dt and gt region
    iscrowd = [int(o["iscrowd"]) for o in gt]

    ious = mask_util.iou(d, g, iscrowd)

    # ####
    # print("id:%s - cat:%d" % (imgId, catId))
    # print("d:", d)
    # print("g:", g)
    # print("ious", ious)
    # ####
    return ious


def compute_AP_dict(ap, aind, mind, eval_dict, iouThr, iouThrList):
    if ap == 1:
        # dimension of precision: [TxRxKxAxM]
        s = eval_dict["precision"]
        # IoU
        if iouThr is not None:
            t = np.where(iouThr == iouThrList)[0]
            s = s[t]
        s = s[:, :, :, aind, mind]
    else:
        # dimension of recall: [TxKxAxM]
        s = eval_dict["recall"]
        if iouThr is not None:
            t = np.where(iouThr == iouThrList)[0]
            s = s[t]
        s = s[:, :, aind, mind]

    if len(s[s > -1]) == 0:
        mean_s = -1
    else:
        mean_s = np.mean(s[s > -1])

    per_class = []
    for i in range(s.shape[2]):
        si = s[:, :, i]
        per_class += [np.mean(si[si > -1])]

    if ap == 1:
        result_dict = {"AP%s" % (iouThr * 100): np.array(per_class), "mAP%s" % (iouThr * 100): mean_s}
    else:
        result_dict = {"RC%s" % (iouThr * 100): np.array(per_class), "mRC%s" % (iouThr * 100): mean_s}

    return result_dict


def evaluateImg(gt_dict, dt_dict, ious, imgId, catId, aRng, maxDet, iouThrs):
    """
    perform evaluation for single category and image
    :return: dict (single image results)
    """
    # p = self.params
    key = (imgId, catId)
    if key in gt_dict:
        gt = gt_dict[key]
    else:
        gt = []

    if key in dt_dict:
        dt = dt_dict[key]
    else:
        dt = []

    if len(gt) == 0 and len(dt) == 0:
        return None

    for g in gt:
        if g["area"] < aRng[0] or g["area"] > aRng[1]:
            g["_ignore"] = 1
        else:
            g["_ignore"] = 0

    # sort dt highest score first, sort gt ignore last
    gtind = np.argsort([g["_ignore"] for g in gt], kind="mergesort")
    gt = [gt[i] for i in gtind]
    dtind = np.argsort([-d["score"] for d in dt], kind="mergesort")
    dt = [dt[i] for i in dtind[0:maxDet]]
    iscrowd = [int(o["iscrowd"]) for o in gt]
    # load computed ious
    ious = ious[(imgId, catId)][:, gtind] if len(ious[(imgId, catId)]) > 0 else ious[(imgId, catId)]

    T = len(iouThrs)
    G = len(gt)
    D = len(dt)
    gtm = np.zeros((T, G))
    dtm = np.zeros((T, D))
    gtIg = np.array([g["_ignore"] for g in gt])
    dtIg = np.zeros((T, D))
    if not len(ious) == 0:
        for tind, t in enumerate(iouThrs):
            for dind, d in enumerate(dt):
                # information about best match so far (m=-1 -> unmatched)
                iou = min([t, 1 - 1e-10])
                m = -1
                for gind, g in enumerate(gt):
                    # if this gt already matched, and not a crowd, continue
                    if gtm[tind, gind] > 0 and not iscrowd[gind]:
                        continue
                    # if dt matched to reg gt, and on ignore gt, stop
                    if m > -1 and gtIg[m] == 0 and gtIg[gind] == 1:
                        break
                    # continue to next gt unless better match made
                    if ious[(dind, gind)] < iou:
                        continue
                    # if match successful and best so far, store appropriately
                    iou = ious[(dind, gind)]
                    m = gind
                # if match made store id of match for both dt and gt
                if m == -1:
                    continue
                dtIg[tind, dind] = gtIg[m]
                dtm[tind, dind] = gt[m]["id"]
                gtm[tind, m] = d["id"]
    # set unmatched detections outside of area range to ignore
    a = np.array([d["area"] < aRng[0] or d["area"] > aRng[1] for d in dt]).reshape((1, len(dt)))
    dtIg = np.logical_or(dtIg, np.logical_and(dtm == 0, np.repeat(a, T, 0)))
    # store results for given image and category
    return {
        "image_id": imgId,
        "category_id": catId,
        "aRng": aRng,
        "maxDet": maxDet,
        "dtIds": [d["id"] for d in dt],
        "gtIds": [g["id"] for g in gt],
        "dtMatches": dtm,
        "gtMatches": gtm,
        "dtScores": [d["score"] for d in dt],
        "gtIgnore": gtIg,
        "dtIgnore": dtIg,
    }


def accumulate(evalImgs, imgIds, catIds, iouThrs, recThrs, areaRng, maxDets):
    """
    Accumulate per image evaluation results and store the result in self.eval
    :param p: input params for evaluation
    :return: None
    """
    # print('Accumulating evaluation results...')

    # allows input customized parameters

    T = len(iouThrs)
    R = len(recThrs)
    K = len(catIds)
    A = len(areaRng)
    M = len(maxDets)

    precision = -np.ones((T, R, K, A, M))  # -1 for the precision of absent categories
    recall = -np.ones((T, K, A, M))
    scores = -np.ones((T, R, K, A, M))

    # create dictionary for future indexing
    catIds = catIds
    setK = set(catIds)
    setA = set(map(tuple, areaRng))
    setM = set(maxDets)
    setI = set(imgIds)
    # get inds to evaluate
    k_list = [n for n, k in enumerate(catIds) if k in setK]
    m_list = [m for n, m in enumerate(maxDets) if m in setM]
    a_list = [n for n, a in enumerate(map(lambda x: tuple(x), areaRng)) if a in setA]
    i_list = [n for n, i in enumerate(imgIds) if i in setI]
    I0 = len(imgIds)
    A0 = len(areaRng)
    # retrieve E at each category, area range, and max number of detections
    for k, k0 in enumerate(k_list):
        Nk = k0 * A0 * I0
        for a, a0 in enumerate(a_list):
            Na = a0 * I0
            for m, maxDet in enumerate(m_list):
                E = [evalImgs[Nk + Na + i] for i in i_list]
                E = [e for e in E if not e is None]
                if len(E) == 0:
                    continue
                dtScores = np.concatenate([e["dtScores"][0:maxDet] for e in E])

                # different sorting method generates slightly different results.
                # mergesort is used to be consistent as Matlab implementation.
                inds = np.argsort(-dtScores, kind="mergesort")
                dtScoresSorted = dtScores[inds]

                dtm = np.concatenate([e["dtMatches"][:, 0:maxDet] for e in E], axis=1)[:, inds]
                dtIg = np.concatenate([e["dtIgnore"][:, 0:maxDet] for e in E], axis=1)[:, inds]
                gtIg = np.concatenate([e["gtIgnore"] for e in E])
                npig = np.count_nonzero(gtIg == 0)
                if npig == 0:
                    continue
                tps = np.logical_and(dtm, np.logical_not(dtIg))
                fps = np.logical_and(np.logical_not(dtm), np.logical_not(dtIg))

                tp_sum = np.cumsum(tps, axis=1).astype(dtype=np.float)
                fp_sum = np.cumsum(fps, axis=1).astype(dtype=np.float)
                for t, (tp, fp) in enumerate(zip(tp_sum, fp_sum)):
                    tp = np.array(tp)
                    fp = np.array(fp)
                    nd = len(tp)
                    rc = tp / npig
                    pr = tp / (fp + tp + np.spacing(1))
                    q = np.zeros((R,))
                    ss = np.zeros((R,))

                    if nd:
                        recall[t, k, a, m] = rc[-1]
                    else:
                        recall[t, k, a, m] = 0

                    # numpy is slow without cython optimization for accessing elements
                    # use python array gets significant speed improvement
                    pr = pr.tolist()
                    q = q.tolist()

                    for i in range(nd - 1, 0, -1):
                        if pr[i] > pr[i - 1]:
                            pr[i - 1] = pr[i]

                    inds = np.searchsorted(rc, recThrs, side="left")
                    try:
                        for ri, pi in enumerate(inds):
                            q[ri] = pr[pi]
                            ss[ri] = dtScoresSorted[pi]
                    except Exception:
                        pass
                    precision[t, :, k, a, m] = np.array(q)
                    scores[t, :, k, a, m] = np.array(ss)
    eval_dict = {
        "counts": [T, R, K, A, M],
        # 'date': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
        "precision": precision,
        "recall": recall,
        "scores": scores,
    }

    return eval_dict


def targets_to_ann_list(images, targets):
    ann_list = []
    ann_id = 0

    img, targets = images[0], targets[0]
    image_id = targets["image_id"].item()
    img_dict = {}
    img_dict["id"] = image_id
    img_dict["height"] = img.shape[-2]
    img_dict["width"] = img.shape[-1]
    # dataset['images'].append(img_dict)
    bboxes = targets["boxes"]
    bboxes[:, 2:] -= bboxes[:, :2]
    bboxes = bboxes.tolist()
    labels = targets["labels"].tolist()
    areas = targets["area"].tolist()
    iscrowd = targets["iscrowd"].tolist()
    if "masks" in targets:
        masks = targets["masks"]
        # make masks Fortran contiguous for coco_mask
        masks = masks.permute(0, 2, 1).contiguous().permute(0, 2, 1)
    if "keypoints" in targets:
        keypoints = targets["keypoints"]
        keypoints = keypoints.reshape(keypoints.shape[0], -1).tolist()
    num_objs = len(bboxes)
    for i in range(num_objs):
        ann = {}
        ann["image_id"] = image_id
        ann["bbox"] = bboxes[i]
        ann["category_id"] = labels[i]
        # categories.add(labels[i])
        ann["score"] = 1
        ann["area"] = areas[i]
        ann["iscrowd"] = iscrowd[i]
        ann["id"] = ann_id
        if "masks" in targets:
            ann["segmentation"] = mask_util.encode(masks[i].numpy())
        if "keypoints" in targets:
            ann["keypoints"] = keypoints[i]
            ann["num_keypoints"] = sum(k != 0 for k in keypoints[i][2::3])

        ann_list += [ann]
        ann_id += 1

    return ann_list


def preds_to_ann_list(preds_dict, mask_void=None):
    ann_list = []
    for original_id, prediction in preds_dict.items():
        if len(prediction) == 0:
            continue

        scores = prediction["scores"]
        labels = prediction["labels"]
        masks = prediction["masks"]

        masks = masks > 0.5

        scores = prediction["scores"].tolist()
        labels = prediction["labels"].tolist()

        rles = []
        for mask in masks:
            binmask = np.array(mask[0, :, :, np.newaxis], dtype=np.uint8, order="F")
            if mask_void is not None:
                ind = mask_void[0][:, :, None].numpy() == 1
                binmask[ind] = 0
            rle = mask_util.encode(binmask)[0]
            rles += [rle]

        for rle in rles:
            rle["counts"] = rle["counts"].decode("utf-8")

        boxes = prediction["boxes"]
        boxes = xyxy_to_xywh(boxes).tolist()

        ann_list.extend(
            [
                {
                    "image_id": original_id,
                    "category_id": labels[k],
                    "bbox": boxes[k],
                    "segmentation": rle,
                    "score": scores[k],
                }
                for k, rle in enumerate(rles)
            ]
        )
    return ann_list


def targets_to_ann_list(images, targets):
    ann_list = []
    ann_id = 0

    img, targets = images[0], targets[0]
    image_id = targets["image_id"].item()
    img_dict = {}
    img_dict["id"] = image_id
    img_dict["height"] = img.shape[-2]
    img_dict["width"] = img.shape[-1]
    # dataset['images'].append(img_dict)
    bboxes = targets["boxes"]
    bboxes[:, 2:] -= bboxes[:, :2]
    bboxes = bboxes.tolist()
    labels = targets["labels"].tolist()
    areas = targets["area"].tolist()
    iscrowd = targets["iscrowd"].tolist()
    if "masks" in targets:
        masks = targets["masks"]
        # make masks Fortran contiguous for coco_mask
        masks = masks.permute(0, 2, 1).contiguous().permute(0, 2, 1)
    if "keypoints" in targets:
        keypoints = targets["keypoints"]
        keypoints = keypoints.reshape(keypoints.shape[0], -1).tolist()
    num_objs = len(bboxes)
    for i in range(num_objs):
        ann = {}
        ann["image_id"] = image_id
        ann["bbox"] = bboxes[i]
        ann["category_id"] = labels[i]
        # categories.add(labels[i])
        ann["score"] = 1
        ann["area"] = areas[i]
        ann["iscrowd"] = iscrowd[i]
        ann["id"] = ann_id
        if "masks" in targets:
            ann["segmentation"] = mask_util.encode(masks[i].numpy())
        if "keypoints" in targets:
            ann["keypoints"] = keypoints[i]
            ann["num_keypoints"] = sum(k != 0 for k in keypoints[i][2::3])

        ann_list += [ann]
        ann_id += 1

    return ann_list


def bbox_yxyx_to_ann_list(bbox_yxyx, image_id=None, category_list=None, score_list=None):
    bbox_xywh = yxyx_to_xywh(bbox_yxyx)

    return bbox_to_ann_list(bbox_xywh, image_id=image_id, category_list=category_list, score_list=score_list)


def bbox_xyxy_to_ann_list(bbox_xyxy, image_id=None, category_list=None, score_list=None):
    bbox_xywh = xyxy_to_xywh(bbox_xyxy)

    return bbox_to_ann_list(bbox_xywh, image_id=image_id, category_list=category_list, score_list=score_list)


def bbox_to_ann_list(bbox_xywh, image_id=None, category_list=None, score_list=None):
    ann_list = []
    _, d = bbox_xywh.shape
    for i in range(bbox_xywh.shape[0]):
        bbox = bbox_xywh[i]
        x, y, w, h = bbox

        ann_list += [
            {
                "iscrowd": 0,
                "bbox": [int(x), int(y), int(w), int(h)],
                "area": int(h * w),
                "image_id": image_id or 0,
                "category_id": category_list or 1,
                "height": int(h),
                "width": int(w),
                "score": score_list or 1,
            }
        ]

    return ann_list


def yxyx_to_xywh(boxes):
    ymin, xmin, ymax, xmax = boxes.unbind(1)
    return torch.stack((xmin, ymin, xmax - xmin, ymax - ymin), dim=1)


def xyxy_to_xywh(boxes):
    xmin, ymin, xmax, ymax = boxes.unbind(1)
    return torch.stack((xmin, ymin, xmax - xmin, ymax - ymin), dim=1)



from setuptools import setup
import setuptools

setup(
    name="haven-ai",
    version="0.6.7",
    description="Manage large-scale experiments",
    url="https://github.com/haven-ai/haven-ai",
    maintainer="Issam Laradji",
    maintainer_email="issam.laradji@gmail.com",
    license="MIT",
    scripts=["./scripts/haven"],
    packages=setuptools.find_packages(),
    zip_safe=False,
    python_requires=">=3.5",
    install_requires=[
        "requests>=0.0",
        "ipywidgets>=0.0",
        "tqdm>=0.0",
        "matplotlib>=0.0",
        "numpy>=0.0",
        "opencv-python-headless>=0.0",
        "pandas>=0.0",
        "Pillow>=0.0",
        "scikit-image>=0.0",
        "scikit-learn>=0.0",
        "scipy>=0.0",
        "sklearn>=0.0",
        "torch>=0.0",
        "torchvision>=0.0",
        "notebook >= 4.0",
        "jupyter>=0.0",
    ],
),



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu
from .. import haven_share as hd

import os
import pprint
import json
import copy
import pprint
import pandas as pd
from . import widgets as wdg

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except Exception:
    print("widgets not available...")


def images_tab(self, output):
    db = self

    if db.vars.get("legend_list") is None:
        db.vars["legend_list"] = hu.get_diff_hparam(db.rm.exp_list)
    w_legend = wdg.SelectMultiple(header="Legend:", options=db.rm.exp_params, db_vars=db.vars, var="legend_list")

    w_n_exps = wdg.Text("n_exps:", default="3", type="int", db_vars=db.vars, var="n_exps")
    w_n_images = wdg.Text("n_images:", default="5", type="int", db_vars=db.vars, var="n_images")
    w_figsize = wdg.Text("figsize:", default="(10,5)", type="tuple", db_vars=db.vars, var="figsize")
    w_dirname = wdg.Text("dirname:", default="images", type="str", db_vars=db.vars, var="dirname")

    bdownload = widgets.Button(description="Download Images", layout=self.layout_button)
    bdownload_out = widgets.Output(layout=self.layout_button)
    bdownload_zip = widgets.Button(description="Download Images zipped", layout=self.layout_button)
    bdownload_zip_out = widgets.Output(layout=self.layout_button)
    brefresh = widgets.Button(description="Display Images")
    button = widgets.VBox(
        [
            widgets.HBox(
                [
                    w_legend.get_widget(),
                    w_n_exps.get_widget(),
                    w_n_images.get_widget(),
                    w_figsize.get_widget(),
                    w_dirname.get_widget(),
                ]
            ),
            widgets.HBox([brefresh, bdownload, bdownload_out, bdownload_zip, bdownload_zip_out]),
        ]
    )

    output_plot = widgets.Output()

    with output:
        display(button)
        display(output_plot)

    def on_clicked(b):
        output_plot.clear_output()
        with output_plot:
            self.update_rm()

            self.rm_original.fig_image_list = self.rm.get_images(
                legend_list=w_legend.update(),
                n_images=w_n_images.update(),
                n_exps=w_n_exps.update(),
                figsize=w_figsize.update(),
                dirname=w_dirname.update(),
            )
            show_inline_matplotlib_plots()

    brefresh.on_click(on_clicked)

    def on_download_clicked(b):
        fname = "images"
        from matplotlib.backends.backend_pdf import PdfPages
        import matplotlib.pyplot as plt

        pp = PdfPages(fname)
        for fig in self.rm_original.fig_image_list:
            fig.savefig(pp, format="pdf")
        pp.close()

        bdownload_out.clear_output()

        with bdownload_out:
            display(FileLink(fname, result_html_prefix="Download: "))

    def on_download_clicked_zip(b):
        fname = "results.zip"
        bdownload_zip_out.clear_output()

        with bdownload_zip_out:
            import zipfile, glob

            exp_id_list = [hu.hash_dict(exp_dict) for exp_dict in self.rm.exp_list]
            zipf = zipfile.ZipFile(fname, "w", zipfile.ZIP_DEFLATED)
            for exp_id in exp_id_list:
                abs_path_list = glob.glob(os.path.join(self.rm.savedir_base, exp_id, "images", "*"))
                for abs_path in abs_path_list:
                    # weq
                    iname = os.path.split(abs_path)[-1]
                    rel_path = f"{exp_id}_{iname}"
                    zipf.write(abs_path, rel_path)
            zipf.close()

            # self.rm.to_zip(savedir_base="", fname=fname, fname_list=self.vars["fname_list"])
        bdownload_zip_out.clear_output()
        with bdownload_zip_out:
            display("%d exps zipped." % len(self.rm.exp_list))
        display(FileLink(fname, result_html_prefix="Download: "))

    bdownload.on_click(on_download_clicked)
    bdownload_zip.on_click(on_download_clicked_zip)



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu
from .. import haven_share as hd
from . import widgets as wdg

import os
import pprint
import json
import copy
import pprint
import pandas as pd


try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except Exception:
    print("widgets not available...")


def share_tab(self, output):
    with output:
        ldownload = widgets.Label(
            value="Get 'score_list.pkl' and 'exp_dict' for each experiment.",
            #    layout=self.layout_button
        )
        bdownload = widgets.Button(description="Download Results", layout=self.layout_button)

        bdownload_out = widgets.Output(layout=self.layout_button)

        # bdownload_dropbox = widgets.Button(description="Upload to Dropbox",
        #                         layout=self.layout_button)

        # bdownload_out_dropbox  = widgets.Output(layout=self.layout_button)

        l_fname_list = widgets.Text(
            value=str(self.vars.get("fname_list", "")),
            layout=self.layout_dropdown,
            description="fname_list:",
            disabled=False,
        )

        # l_dropbox_path = widgets.Text(
        #     value=str(self.vars.get('dropbox_path', '/shared')),
        #     description='dropbox_path:',
        #     layout=self.layout_dropdown,
        #     disabled=False
        #         )
        # l_access_token_path = widgets.Text(
        #     value=str(self.vars.get('access_token', '')),
        #     description='access_token:',
        #     layout=self.layout_dropdown,
        #     disabled=False
        #         )
        def on_upload_clicked(b):
            fname = "results.zip"
            bdownload_out_dropbox.clear_output()
            self.vars["fname_list"] = hu.get_list_from_str(l_fname_list.value)
            self.vars["dropbox_path"] = l_dropbox_path.value
            self.vars["access_token"] = l_access_token_path.value
            with bdownload_out_dropbox:
                self.rm.to_zip(
                    savedir_base="",
                    fname=fname,
                    fname_list=self.vars["fname_list"],
                    dropbox_path=self.vars["dropbox_path"],
                    access_token=self.vars["access_token"],
                )

            os.remove("results.zip")
            display("result.zip sent to dropbox at %s." % self.vars["dropbox_path"])

        def on_download_clicked(b):
            fname = "results.zip"
            bdownload_out.clear_output()
            self.vars["fname_list"] = hu.get_list_from_str(l_fname_list.value)

            with bdownload_out:
                self.rm.to_zip(savedir_base="", fname=fname, fname_list=self.vars["fname_list"])
            bdownload_out.clear_output()
            with bdownload_out:
                display("%d exps zipped." % len(self.rm.exp_list))
            display(FileLink(fname, result_html_prefix="Download: "))

        bdownload.on_click(on_download_clicked)
        bdownload_zip = widgets.VBox([bdownload, bdownload_out])
        # bdownload_dropbox.on_click(on_upload_clicked)
        # bdownload_dropbox_vbox = widgets.VBox([ bdownload_dropbox, bdownload_out_dropbox])
        display(
            widgets.VBox(
                [
                    # widgets.HBox([l_fname_list, l_dropbox_path, l_access_token_path]),
                    ldownload,
                    widgets.HBox([bdownload_zip]),
                ]
            )
        )
